from nipype.interfaces.fsl import BET, FLIRT, RobustFOV, ApplyXFM, ApplyMask
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.CustomDcm2niix import CustomDcm2niix
from swane.nipype_pipeline.nodes.ForceOrient import ForceOrient
from nipype import Node
from nipype.interfaces.utility import IdentityInterface
from configparser import SectionProxy


def linear_reg_workflow(
    name: str,
    dicom_dir: str,
    config: SectionProxy,
    base_dir: str = "/",
    is_volumetric: bool = True,
    is_partial_coverage: bool = False,
) -> CustomWorkflow:
    """
    Transforms input images in a reference space through a linear registration.

    Parameters
    ----------
    name : str
        The workflow name.
    dicom_dir : path
        The file path of the DICOM files.
    config: SectionProxy
        workflow settings.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".
    is_volumetric : bool, optional
        True if input is 3D. The default is True.
    is_partial_coverage : bool, optional
        True if series only includes brain partially. The default is False.

    Input Node Fields
    ----------
    reference : path
        The reference image for the registration.
    output_name : str
        The name for registered file.
    brain_mask : path
        The brain mask image. Only needed if is_partial_coverage is True.

    Returns
    -------
    workflow : CustomWorkflow
        The linear registration workflow.

    Output Node Fields
    ----------
    registered_file : string
        Output file in T13D reference space.
    betted_registered_file : string
        Output betted file in T13D reference space.
    out_matrix_file : path
        Linear registration matrix to T13D reference space.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(
        IdentityInterface(fields=["reference", "output_name", "brain_mask"]),
        name="inputnode",
    )

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=["registered_file", "betted_registered_file" "out_matrix_file"]
        ),
        name="outputnode",
    )

    # NODE 1: Conversion dicom -> nifti
    conversion = Node(CustomDcm2niix(), name="%s_conv" % name)
    conversion.inputs.source_dir = dicom_dir
    conversion.inputs.bids_format = False
    conversion.inputs.out_filename = name
    conversion.inputs.name_conflicts = 1
    conversion.inputs.merge_imgs = 2

    # NODE 2: Orienting in radiological convention
    reorient = Node(ForceOrient(), name="%s_reorient" % name)
    workflow.connect(conversion, "converted_files", reorient, "in_file")

    # NODE 3: Crop neck
    robustfov = Node(RobustFOV(), name="%s_robustfov" % name)
    workflow.connect(reorient, "out_file", robustfov, "in_file")

    def get_betted_name(basename):
        return "r-%s_brain.nii.gz" % basename

    def get_unbetted_name(basename):
        return "r-%s.nii.gz" % basename

    if is_partial_coverage:
        # NODE 4a: Linear registration to reference space
        flirt_2_ref = Node(FLIRT(), name="%s_2_ref" % name)
        flirt_2_ref.long_name = "%s to reference space"
        flirt_2_ref.inputs.out_matrix_file = "%s_2_ref.mat" % name

        flirt_2_ref.inputs.cost = "mutualinfo"
        flirt_2_ref.inputs.searchr_x = [-40, 40]
        flirt_2_ref.inputs.searchr_y = [-40, 40]
        flirt_2_ref.inputs.searchr_z = [-40, 40]
        flirt_2_ref.inputs.dof = 6
        flirt_2_ref.inputs.interp = "trilinear"

        workflow.connect(robustfov, "out_roi", flirt_2_ref, "in_file")
        workflow.connect(
            inputnode, ("output_name", get_unbetted_name), flirt_2_ref, "out_file"
        )
        workflow.connect(inputnode, "reference", flirt_2_ref, "reference")

        # NODE 5a: Apply brain mask
        brain_masking = Node(ApplyMask(), name="%s_brain_mask" % name)
        brain_masking.long_name = "Brain %s"
        workflow.connect(flirt_2_ref, "out_file", brain_masking, "in_file")
        workflow.connect(
            inputnode, ("output_name", get_betted_name), brain_masking, "out_file"
        )
        workflow.connect(inputnode, "brain_mask", brain_masking, "mask_file")

        workflow.connect(flirt_2_ref, "out_file", outputnode, "registered_file")
        workflow.connect(
            brain_masking, "out_file", outputnode, "betted_registered_file"
        )
        workflow.connect(flirt_2_ref, "out_matrix_file", outputnode, "out_matrix_file")

    else:
        # NODE 4b: Scalp removal
        bet = Node(BET(), "%s_BET" % name)
        if config is not None:
            bet.inputs.frac = config.getfloat_safe("bet_thr")
        if config is not None and config.getboolean_safe("bet_bias_correction"):
            bet.inputs.reduce_bias = True
        else:
            bet.inputs.robust = True
        bet.inputs.mask = True
        workflow.connect(robustfov, "out_roi", bet, "in_file")

        # NODE 5b: Linear registration to reference space
        flirt_2_ref = Node(FLIRT(), name="%s_brain_2_ref" % name)
        flirt_2_ref.long_name = "%s to reference space"
        flirt_2_ref.inputs.out_matrix_file = "%s_2_ref.mat" % name

        if is_volumetric:
            flirt_2_ref.inputs.cost = "mutualinfo"
            flirt_2_ref.inputs.searchr_x = [-90, 90]
            flirt_2_ref.inputs.searchr_y = [-90, 90]
            flirt_2_ref.inputs.searchr_z = [-90, 90]
            flirt_2_ref.inputs.dof = 6
            flirt_2_ref.inputs.interp = "trilinear"

        workflow.connect(bet, "out_file", flirt_2_ref, "in_file")
        workflow.connect(
            inputnode, ("output_name", get_betted_name), flirt_2_ref, "out_file"
        )
        workflow.connect(inputnode, "reference", flirt_2_ref, "reference")

        # NODE 6b: Linear trasformation of unbetted series to reference space
        unbet_flirt = Node(ApplyXFM(), name="%s_2_ref" % name)
        unbet_flirt.long_name = "%s to reference space"

        if is_volumetric:
            unbet_flirt.inputs.cost = "mutualinfo"
            unbet_flirt.inputs.searchr_x = [-90, 90]
            unbet_flirt.inputs.searchr_y = [-90, 90]
            unbet_flirt.inputs.searchr_z = [-90, 90]
            unbet_flirt.inputs.dof = 6
            unbet_flirt.inputs.interp = "trilinear"

        workflow.connect(robustfov, "out_roi", unbet_flirt, "in_file")
        workflow.connect(flirt_2_ref, "out_matrix_file", unbet_flirt, "in_matrix_file")
        workflow.connect(
            inputnode, ("output_name", get_unbetted_name), unbet_flirt, "out_file"
        )
        workflow.connect(inputnode, "reference", unbet_flirt, "reference")

        workflow.connect(flirt_2_ref, "out_file", outputnode, "betted_registered_file")
        workflow.connect(unbet_flirt, "out_file", outputnode, "registered_file")
        workflow.connect(flirt_2_ref, "out_matrix_file", outputnode, "out_matrix_file")

    return workflow



import time
from functools import wraps


def debug(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        print(f"Calling {func.__name__} with args: {args} and kwargs: {kwargs}")
        result = func(*args, **kwargs)
        print(f"{func.__name__} returned: {result}")
        return result

    return wrapper


def timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"Execution time of {func.__name__}: {end - start} seconds")
        return result

    return wrapper



import pydicom


class DicomSeries:
    def __init__(self):
        self.dicom_locs = []
        self.frames = 0
        self.is_multi_frame = False
        self.multi_frame_loc = None
        self.first_position = None
        self.volumes = 1
        self.description = "Not named"
        self.modality = None
        self.classification = "Not classified"
        self.ds = None

    def add_dicom_loc(self, dicom_loc, is_multi_frame, slice_loc, ds=None):
        if dicom_loc not in self.dicom_locs:
            self.dicom_locs.append(dicom_loc)
            if is_multi_frame:
                self.is_multi_frame = is_multi_frame
                self.multi_frame_loc = dicom_loc
                # Save dicom set for multi frame series to avoid long re-read in refine_frame_number loop at scan ending
                self.ds = ds
            else:
                self.frames += 1
                if self.first_position is None:
                    self.first_position = slice_loc
                elif self.first_position == slice_loc:
                    self.volumes += 1

    def refine_frame_number(self):
        if self.is_multi_frame and self.multi_frame_loc is not None:
            if self.ds is None:
                self.ds = pydicom.dcmread(self.multi_frame_loc, force=True)
            for i, frame_group in enumerate(self.ds.PerFrameFunctionalGroupsSequence):
                if i == 0:
                    self.first_position = frame_group.PlanePositionSequence
                    self.volumes = 1
                    self.frames = int(self.ds.NumberOfFrames)
                elif self.first_position == frame_group.PlanePositionSequence:
                    self.volumes += 1
            # Free memory from potentially large dicom set we don't need any more
            self.ds = None
            del self.ds
        elif self.frames < 10:
            ds = pydicom.dcmread(self.dicom_locs[0], force=True)
            if not hasattr(ds, "ImageType") or "MOSAIC" not in ds.ImageType:
                self.frames = 0


class DicomSubject:
    def __init__(self, subject_id: str, subject_name: str):
        self.subject_id = subject_id
        self.subject_name = subject_name
        self.studies = {}

    def add_study(self, study_instance_uid):
        if study_instance_uid not in self.studies:
            self.studies[study_instance_uid] = {}

    def add_series(self, study_instance_uid, series_number):
        if study_instance_uid not in self.studies:
            raise Exception("Study " + study_instance_uid + " not found")
        if series_number not in self.studies[study_instance_uid]:
            self.studies[study_instance_uid][series_number] = DicomSeries()
        return self.studies[study_instance_uid][series_number]

    def get_series_list(
        self, study_instance_uid: pydicom.uid.UID
    ) -> list[pydicom.valuerep.IS]:
        if study_instance_uid not in self.studies:
            return []
        else:
            return list(self.studies[study_instance_uid].keys())

    def get_series(
        self, study_instance_uid: pydicom.uid.UID, series_number
    ) -> DicomSeries:
        if study_instance_uid not in self.studies:
            return []
        elif series_number not in self.studies[study_instance_uid]:
            return []
        else:
            return self.studies[study_instance_uid][series_number]


class DicomTree:
    def __init__(self, dicom_dir: str):
        self.dicom_subjects = {}
        self.dicom_dir = dicom_dir

    def add_subject(self, subject_id, subject_name):
        if subject_id not in self.dicom_subjects:
            self.dicom_subjects[subject_id] = DicomSubject(subject_id, subject_name)

    def add_study(self, subject_id, study_instance_uid):
        if subject_id not in self.dicom_subjects:
            raise Exception("Subject " + subject_id + " not found")
        self.dicom_subjects[subject_id].add_study(study_instance_uid)

    def add_series(self, subject_id, study_instance_uid, series_number) -> DicomSeries:
        if subject_id not in self.dicom_subjects:
            raise Exception("Subject " + subject_id + " not found")
        return self.dicom_subjects[subject_id].add_series(
            study_instance_uid, series_number
        )

    def get_subject_list(self):
        return list(self.dicom_subjects.keys())

    def get_studies_list(self, subject: str) -> list[pydicom.uid.UID]:
        """
        Extract from dicom search the studies of specified subject and return their study_id

        Parameters
        ----------
        subject: str
            The subject id

        Returns
        -------
            A list of study_id
        """
        if subject not in self.dicom_subjects:
            return []
        return list(self.dicom_subjects[subject].studies.keys())

    def get_series_list(
        self, subject: str, study_instance_uid: pydicom.uid.UID
    ) -> list[pydicom.valuerep.IS]:
        """
        Extract from dicom search the series of a specified studies of specified subject and return their series_id

        Parameters
        ----------
        subject: str
            The subject id
        study_instance_uid: pydicom.uid.UID
            The study id

        Returns
        -------
            A list of series_id
        """
        if subject not in self.dicom_subjects:
            return []
        return self.dicom_subjects[subject].get_series_list(study_instance_uid)

    def get_series(
        self, subject: str, study_instance_uid: pydicom.uid.UID, series_number
    ) -> DicomSeries:
        if subject not in self.dicom_subjects:
            return None
        return self.dicom_subjects[subject].get_series(
            study_instance_uid, series_number
        )



import shutil
import platform


def is_command_available(command: str) -> bool:
    """
    Check if a command exists in the system PATH.

    :param command: Name of the command to check
    :return: True if the command exists, False otherwise
    """
    return shutil.which(command) is not None


def get_os_type() -> str:
    """
    Get the operating system type.

    :return: 'mac' if macOS, 'linux' if Linux, 'other' otherwise
    """
    system = platform.system().lower()
    if system == "darwin":
        return "mac"
    elif system == "linux":
        return "linux"
    else:
        return "other"


def is_mac() -> bool:
    """
    Check if the operating system is macOS.
    """
    return get_os_type() == "mac"


def is_linux() -> bool:
    """
    Check if the operating system is Linux.
    """
    return get_os_type() == "linux"



class NodeListEntry:
    """
    Utility class used to get generic info of a Nipype Node.
    long_name: explicit name of the Node.
    node_list: if a node is a Workflow, the list of the its subnodes info stored in a list of NodeListEntry objects.
    node_holder: CustomTreeWidgetItem object to display the Node status in UI.

    """

    long_name = None
    node_list = {}
    node_holder = None



import os
import psutil


def last_pid_is_running(last_pid: int) -> bool:
    """

    Parameters
    ----------
    last_pid: int
        The last previous application launch process id

    Returns
    -------
    True if a process with PID=lòast_pid is running
    """
    try:
        return last_pid != os.getpid() and psutil.Process(last_pid)
    except (psutil.NoSuchProcess, ValueError):
        pass
    return False



from PySide6.QtWidgets import (
    QMainWindow,
    QMessageBox,
    QFileDialog,
    QInputDialog,
    QStyle,
    QLineEdit,
    QTabWidget,
    QGridLayout,
    QLabel,
    QSizePolicy,
    QSpacerItem,
    QWidget,
    QTabBar,
    QDialog,
    QPushButton,
    QStyleOptionButton,
)
from PySide6.QtGui import QAction, QIcon, QPixmap, QFont, QCloseEvent
from PySide6.QtCore import QCoreApplication, Qt, QThreadPool
from PySide6.QtSvgWidgets import QSvgWidget
import os
from swane.utils.DependencyManager import (
    DependencyManager,
    Dependence,
    DependenceStatus,
)
from swane.ui.SubjectTab import SubjectTab
from swane.ui.PreferencesWindow import PreferencesWindow
import swane_supplement
from swane import __version__, EXIT_CODE_REBOOT, strings
from swane.workers.UpdateCheckWorker import UpdateCheckWorker
from swane.utils.Subject import Subject, SubjectRet
from swane.config.ConfigManager import ConfigManager
from swane.config.config_enums import GlobalPrefCategoryList


class MainWindow(QMainWindow):
    """
    Custom implementation of PySide QMainWindow to define SWANe GUI.

    """

    def __init__(self, global_config: ConfigManager):

        super(MainWindow, self).__init__()

        # GUI configuration setting
        self.global_config: ConfigManager = global_config
        self.dependency_manager: DependencyManager = DependencyManager()
        self.global_config.check_dependencies(self.dependency_manager)

        # GUI Icons setting
        self.setWindowIcon(QIcon(QPixmap(swane_supplement.appIcon_file)))
        self.OK_ICON_FILE = swane_supplement.okIcon_file
        self.ERROR_ICON_FILE = swane_supplement.errorIcon_file
        self.WARNING_ICON_FILE = swane_supplement.warnIcon_file
        self.LOADING_MOVIE_FILE = swane_supplement.loadingMovie_file
        self.VOID_SVG_FILE = swane_supplement.voidsvg_file
        self.OK_ICON = QPixmap(self.OK_ICON_FILE)
        self.ERROR_ICON = QPixmap(self.ERROR_ICON_FILE)
        self.WARNING_ICON = QPixmap(self.WARNING_ICON_FILE)
        self.NON_UNICODE_BUTTON_HEIGHT = MainWindow.get_non_unicode_height()

        # subject folder configuration checking
        while self.global_config.get_main_working_directory() == "":
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_choose_working_dir)
            msg_box.exec()
            self.set_main_working_directory()

        # subject folder configuration setting
        os.chdir(self.global_config.get_main_working_directory())

        self.initialize_ui()

        # Check for update
        update_thread = UpdateCheckWorker()
        update_thread.signal.last_available.connect(
            lambda pip_version: self.update_available(pip_version)
        )
        QThreadPool.globalInstance().start(update_thread)

    def update_available(self, pip_version: str):
        """
        Called if UpdateCheckWorker detect a newer version on pip
        Parameters
        ----------
        pip_version: str
            The version of the update available on pip
        """
        msg_box = QMessageBox(parent=self)
        msg_box.setIcon(QMessageBox.Icon.Information)
        msg_box.setText(strings.mainwindow_update_available % pip_version)
        msg_box.exec()

    @staticmethod
    def get_non_unicode_height() -> int:
        """
        Returns
        -------
        The pixel height of a generic label WITHOUT unicode character
        """
        button = QPushButton()
        opt = QStyleOptionButton()
        opt.initFrom(button)
        text_size = button.fontMetrics().size(Qt.TextShowMnemonic, button.text())
        return (
            button.style()
            .sizeFromContents(QStyle.CT_PushButton, opt, text_size, button)
            .height()
        )

    def open_subject_tab(self, subject: Subject, check_dicom_folders: bool = True):
        """
        Load a checked and valid subject folder.

        Parameters
        ----------
        subject : str
            The subject to load in the tab.

        check_dicom_folders : bool
            If True, check for dicom files in the fubject folders. Default is True.

        Returns
        -------
        None.

        """

        this_tab = SubjectTab(
            self.global_config, subject, main_window=self, parent=self.main_tab
        )
        self.subject_tab_array.append(this_tab)

        self.main_tab.addTab(this_tab, os.path.basename(subject.name))
        self.main_tab.setCurrentWidget(this_tab)

        this_tab.load_subject(check_dicom_folders)

    def check_subject_limit(self) -> bool:
        """
        Check if SWANe can open another subject tab without overcome the limit set by configuration.

        Returns
        -------
        bool
            True if SWANe can load another tab, otherwise False.

        """

        max_subjects = self.global_config.get_max_subject_tabs()
        if max_subjects <= 0:
            return True
        if len(self.subject_tab_array) >= max_subjects:
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Warning)
            msg_box.setText(strings.mainwindow_max_subj_error)
            msg_box.exec()
            return False
        return True

    def search_subject_dir(self, button_state: bool = False, folder_path: str = None):
        """
        Try to open a subject directory

        Parameters
        ----------
        button_state: bool
            Not used, passed by QPushButton. Default is False
        folder_path: str
            The folder to load. Default is None

        Returns
        -------
        None.

        """

        # Guard to avoid the opening of subject tabs greater than the maximum allowed
        if not self.check_subject_limit():
            return

        if folder_path is None:
            # Open the directory selection dialog if a path is not provided
            file_dialog = QFileDialog()
            file_dialog.setDirectory(self.global_config.get_main_working_directory())
            folder_path = file_dialog.getExistingDirectory(
                self, strings.mainwindow_select_subj_folder
            )

        subject = Subject(
            self.global_config, dependency_manager=self.dependency_manager
        )

        # Guard to avoid an already loaded subject directory
        for tab in self.subject_tab_array:
            if tab.subject.folder == folder_path:
                msg_box = QMessageBox()
                msg_box.setIcon(QMessageBox.Icon.Warning)
                msg_box.setText(strings.mainwindow_subj_already_loaded_error)
                msg_box.exec()
                return

        subject_load_ret = subject.load(folder_path)

        if subject_load_ret == SubjectRet.ValidFolder:
            self.open_subject_tab(subject)
        elif subject_load_ret == SubjectRet.FolderNotFound:
            # User canceled subject load
            return
        elif subject_load_ret == SubjectRet.PathBlankSpaces:
            # Guard to avoid the opening a directory containing blank spaces
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.setText(strings.mainwindow_subj_folder_with_blank_spaces_error)
            msg_box.exec()
            return
        elif subject_load_ret == SubjectRet.FolderOutsideMain:
            # Guard to avoid the opening of a directory outside the main subject folder
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.setText(strings.mainwindow_subj_folder_outside_workingdir_error)
            msg_box.exec()
            return
        elif subject_load_ret == SubjectRet.InvalidFolderTree:
            # Check if selected folder is a valid subject folder
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Warning)
            msg_box.setText(strings.mainwindow_invalid_folder_error)
            msg_box.exec()

            msg_box2 = QMessageBox()
            msg_box2.setText(strings.mainwindow_force_dir_update)
            msg_box2.setIcon(QMessageBox.Icon.Question)
            msg_box2.setStandardButtons(
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
            )
            msg_box2.button(QMessageBox.StandardButton.Yes).setText("Yes")
            msg_box2.button(QMessageBox.StandardButton.No).setText("No")
            msg_box2.setDefaultButton(QMessageBox.StandardButton.No)
            ret = msg_box2.exec()

            # A subject folder has a predefined folder tree.
            # SWANe recognizes a subject folder checking its subfolders.
            # If a selected folder is not valid, the user may force its conversion into a subject folder.
            if ret == QMessageBox.StandardButton.Yes:
                subject.fix_subject_folder_subtree(folder_path)
                self.search_subject_dir(folder_path=folder_path)
            return

        return

    def get_suggested_subject_name(self) -> str:
        """
        Get a default name for the subject folder based the existing subject folders into the main subject directory.

        Returns
        -------
        str
            The suggested subject folder name.

        """

        import re

        regex = re.compile("^" + self.global_config.get_subjects_prefix() + r"\d+$")
        file_list = []

        for this_dir in os.listdir(self.global_config.get_main_working_directory()):
            if regex.match(this_dir):
                file_list.append(
                    int(this_dir.replace(self.global_config.get_subjects_prefix(), ""))
                )

        if len(file_list) == 0:
            return self.global_config.get_subjects_prefix() + "1"

        return self.global_config.get_subjects_prefix() + str(max(file_list) + 1)

    def choose_new_subject_dir(self):
        """
        Create a new subject folder. The user must specify its name.

        Returns
        -------
        None.

        """

        if not self.check_subject_limit():
            return

        text, ok = QInputDialog.getText(
            self,
            strings.mainwindow_new_subj_title,
            strings.mainwindow_new_subj_name,
            QLineEdit.EchoMode.Normal,
            self.get_suggested_subject_name(),
        )

        if not ok:
            return

        subject_name = str(text).replace(" ", "_")
        subject = Subject(
            self.global_config, dependency_manager=self.dependency_manager
        )
        create_subject_ret = subject.create_new_subject_dir(subject_name)

        if (
            create_subject_ret == SubjectRet.FolderNotFound
            or create_subject_ret == SubjectRet.PathBlankSpaces
        ):
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_new_subj_name_error + subject_name)
            msg_box.exec()
            return
        elif create_subject_ret == SubjectRet.FolderAlreadyExists:
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_subj_exists_error + subject_name)
            msg_box.exec()
            return
        elif create_subject_ret == SubjectRet.ValidFolder:
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_new_subj_created + subject_name)
            msg_box.exec()
            self.open_subject_tab(subject, check_dicom_folders=False)

    def set_main_working_directory(self):
        """
        Generates the OS directory selection dialog to set the default subject folder

        Returns
        -------
        None.

        """

        folder_path = QFileDialog.getExistingDirectory(
            self, strings.mainwindow_choose_working_dir_title
        )

        if not os.path.exists(folder_path):
            return

        if " " in folder_path:
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_working_dir_space_error)
            msg_box.exec()
            return

        self.global_config.set_main_working_directory(os.path.abspath(folder_path))

        os.chdir(folder_path)

    def edit_config(self):
        """
        Open the Global Preferences Window.

        Returns
        -------
        None.

        """

        if self.check_running_workflows():
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_pref_disabled_error)
            msg_box.exec()
            return

        preference_window = PreferencesWindow(
            self.global_config, self.dependency_manager, False
        )
        ret = preference_window.exec()

        if ret == EXIT_CODE_REBOOT:
            self.close()
            QCoreApplication.exit(EXIT_CODE_REBOOT)

        if ret != 0:
            self.reset_workflows()

    def edit_wf_config(self):
        """
        Open the Default Workflow Settings Window.

        Returns
        -------
        None.

        """

        if self.check_running_workflows():
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_pref_disabled_error)
            msg_box.exec()
            return

        wf_preference_window = PreferencesWindow(
            self.global_config, self.dependency_manager, True
        )
        ret = wf_preference_window.exec()

        if ret == -1:
            self.global_config.reset_to_defaults()
            self.edit_wf_config()

    def toggle_shutdown_after_workflow(self):
        """
        Toggle shutdown after woprkflow preference

        Returns
        -------
        None.

        """
        shutdown = self.global_config.getboolean_safe(
            GlobalPrefCategoryList.MAIN, "shutdown"
        )
        self.global_config[GlobalPrefCategoryList.MAIN]["shutdown"] = str(not shutdown)

    def check_running_workflows(self, ignore_subj: Subject = None) -> bool:
        """
        Check if SWANe is executing a workflow in any open subject tab.

        Parameters
        ----------
        ignore_subj : Subject
            The subject tab to be ignored. Default is None

        Returns
        -------
        bool
            True if SWANe is executing a workflow, otherwise False.

        """

        for subj in self.subject_tab_array:
            if subj.subject != ignore_subj and subj.subject.is_workflow_process_alive():
                return True

        return False

    def check_workflow_error(self) -> bool:
        """
        Check if there was a node error in any open subject tab.

        Returns
        -------
        bool
            True if there was any error, otherwise False.

        """

        for subj in self.subject_tab_array:
            if subj.workflow_had_error:
                return True
        return False

    def reset_workflows(self):
        """
        Reset all the generated workflows that are not already running.

        Returns
        -------
        None.

        """

        for subj in self.subject_tab_array:
            subj.reset_workflow()

    def about(self):
        """
        Open the About Window.

        Returns
        -------
        None.

        """

        about_dialog = QDialog(parent=self)
        layout = QGridLayout()

        bold_font = QFont()
        bold_font.setBold(True)
        title_font = QFont()
        title_font.setBold(True)
        title_font.setPointSize(title_font.pointSize() * 1.5)

        label_about1 = QLabel(strings.APPNAME)
        label_about1.setFont(title_font)
        label_about2 = QLabel(strings.app_acronym)
        label_about3 = QLabel("Version: " + __version__)

        label_about4 = QLabel(strings.aboutwindow_wiki_changelog)
        label_about5 = QLabel(strings.aboutwindow_wiki_dependencylist)

        label_about_icon = QLabel()
        icon = QPixmap(swane_supplement.appIcon_file)

        label_about_icon.setPixmap(icon.scaled(60, 60))

        layout.addWidget(label_about1, 0, 1)
        layout.addWidget(label_about2, 1, 1)
        layout.addWidget(label_about3, 2, 1)
        layout.addWidget(label_about4, 3, 1)
        layout.addWidget(label_about5, 4, 1)

        layout.addWidget(label_about_icon, 0, 0, 3, 1)

        about_dialog.setLayout(layout)
        about_dialog.exec()

    def initialize_ui(self):
        """
        Generates the SWANE GUI

        Returns
        -------
        None.

        """

        self.resize(800, 600)
        self.setWindowTitle(strings.APPNAME + " - " + strings.app_acronym)

        self.statusBar().showMessage("")

        # Buttons definition
        button_action = QAction(
            QIcon.fromTheme("document-open"), strings.menu_load_subj, self
        )
        button_action.setStatusTip(strings.menu_load_subj_tip)
        button_action.triggered.connect(self.search_subject_dir)

        button_action2 = QAction(
            QIcon.fromTheme("document-new"), strings.menu_new_subj, self
        )
        button_action2.setStatusTip(strings.menu_new_subj_tip)
        button_action2.triggered.connect(self.choose_new_subject_dir)

        button_action3 = QAction(
            QIcon.fromTheme("application-exit"), strings.menu_exit, self
        )
        button_action3.triggered.connect(self.close)

        button_action4 = QAction(
            QIcon.fromTheme("preferences-other"), strings.menu_pref, self
        )
        button_action4.setStatusTip(strings.menu_pref_tip)
        button_action4.triggered.connect(self.edit_config)

        button_action5 = QAction(
            QIcon.fromTheme("preferences-other"), strings.menu_wf_pref, self
        )
        button_action5.triggered.connect(self.edit_wf_config)

        button_action6 = QAction(strings.menu_shutdown_pref, self)
        button_action6.setCheckable(True)
        button_action6.triggered.connect(self.toggle_shutdown_after_workflow)

        button_action7 = QAction(strings.menu_about, self)
        button_action7.triggered.connect(self.about)

        # Menu definition and population
        menu = self.menuBar()
        menu.setNativeMenuBar(False)
        file_menu = menu.addMenu(strings.menu_file_name)
        file_menu.addAction(button_action)
        file_menu.addAction(button_action2)
        file_menu.addAction(button_action3)
        tool_menu = menu.addMenu(strings.menu_tools_name)
        tool_menu.addAction(button_action4)
        tool_menu.addAction(button_action5)
        tool_menu.addAction(button_action6)
        help_menu = menu.addMenu(strings.menu_help_name)
        help_menu.addAction(button_action7)

        # Tab definition
        self.main_tab = QTabWidget(parent=self)
        self.main_tab.setTabsClosable(True)
        self.main_tab.tabCloseRequested.connect(self.close_subject_tab)
        self.setCentralWidget(self.main_tab)
        self.homeTab = QWidget()
        self.main_tab.addTab(self.homeTab, strings.mainwindow_home_tab_name)

        # Tab closing option disabled
        self.main_tab.tabBar().setTabButton(0, QTabBar.ButtonPosition.LeftSide, None)
        self.main_tab.tabBar().setTabButton(0, QTabBar.ButtonPosition.RightSide, None)

        # Home Tab definition
        self.home_tab_ui()

        self.subject_tab_array = []

        # Link to ChatGPT SWANe Assistant - as a clear clickable QPushButton
        chatgpt_url = (
            "https://chatgpt.com/g/g-68e14421a54c8191a2110a831824f1e9-swaneassistant/"
        )
        chatgpt_text = f"🤖 {strings.mainwindow_chatgpt_title}"

        self.chatgpt_button = QPushButton(chatgpt_text, parent=self)
        self.chatgpt_button.setCursor(Qt.PointingHandCursor)
        self.chatgpt_button.setToolTip(strings.mainwindow_chatgpt_tooltip)
        self.chatgpt_button.setMinimumHeight(28)
        self.chatgpt_button.setContentsMargins(4, 0, 4, 0)
        self.chatgpt_button.setStyleSheet(
            """
            QPushButton {
                color: #ffffff;
                background-color: rgba(0, 102, 204, 0.12);
                padding: 6px 12px;
                border: 1px solid rgba(0, 102, 204, 0.20);
                border-radius: 8px;
                font-weight: 600;
                font-size: 13px;
            }
            QPushButton:hover {
                background-color: rgba(0, 102, 204, 0.18);
            }
            QPushButton:pressed {
                background-color: rgba(0, 102, 204, 0.24);
            }
        """
        )

        def _open_chatgpt():
            import webbrowser

            webbrowser.open(chatgpt_url)

        self.chatgpt_button.clicked.connect(_open_chatgpt)

        # Add to status bar (use addPermanentWidget per preferenza)
        self.statusBar().addPermanentWidget(self.chatgpt_button)

        self.show()

    def close_subject_tab(self, index: int):
        """
        Handle the PySide tab closing event for the subject tab.

        Parameters
        ----------
        index : int
            The subject tab index into the GUI.

        Returns
        -------
        None.

        """

        # Guard to prevent the Home Tab closing
        if index <= 0:
            return

        tab_item = self.main_tab.widget(index)
        if tab_item.subject.is_workflow_process_alive():
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_wf_executing_error_1)
            msg_box.exec()
            return

        tab_item.subject.config.save()

        self.subject_tab_array.remove(tab_item)
        self.main_tab.removeTab(index)

    def closeEvent(self, event: QCloseEvent):
        """
        Prevent the closing of a running workflow tab

        Parameters
        ----------
        event : QCloseEvent
            PySide QCloseEvent.

        Returns
        -------
        None.

        """

        if not self.check_running_workflows():
            event.accept()
        else:
            msg_box = QMessageBox()
            msg_box.setText(strings.mainwindow_wf_executing_error_2)
            msg_box.exec()
            event.ignore()

    def safe_shutdown_after_workflow(self, subject: Subject):
        """
        Shutdown pc if all workflows are completed

        Parameters
        ----------
        subject: Subject
            The subject tab asking to shutdown

        Returns
        -------
        None.

        """

        if (
            not self.check_running_workflows(ignore_subj=subject)
            and not self.check_workflow_error()
        ):
            os.system("systemctl poweroff")
            os.system("osascript -e 'tell app \"System Events\" to shut down'")
            exit()
            pass

    def home_tab_ui(self):
        """
        Generates the Home Tab layout

        Returns
        -------
        None.

        """

        self.home_grid_layout = QGridLayout()

        bold_font = QFont()
        bold_font.setBold(True)
        title_font = QFont()
        title_font.setBold(True)
        title_font.setPointSize(title_font.pointSize() * 1.5)
        x = 0

        label_welcome1 = QLabel(strings.mainwindow_home_label1)
        label_welcome1.setFont(title_font)
        label_welcome2 = QLabel(strings.mainwindow_home_label2)
        label_welcome2.setWordWrap(True)
        label_welcome3 = QLabel(strings.mainwindow_home_label3)
        label_welcome3.setWordWrap(True)
        label_welcome4 = QLabel(strings.mainwindow_home_label4)
        label_welcome4.setFont(bold_font)

        self.home_grid_layout.addWidget(label_welcome1, x, 0, 1, 2)
        x += 1
        self.home_grid_layout.addWidget(label_welcome2, x, 0, 1, 2)
        x += 1
        self.home_grid_layout.addWidget(label_welcome3, x, 0, 1, 2)
        x += 1
        self.home_grid_layout.addWidget(label_welcome4, x, 0, 1, 2)
        x += 1

        # Main window dependency check
        label_main_dep = QLabel(strings.mainwindow_home_label5)
        label_main_dep.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label_main_dep.setFont(bold_font)
        self.home_grid_layout.addWidget(label_main_dep, x, 0, 1, 2)
        x += 1

        x = self.add_home_entry(self.dependency_manager.dcm2niix, x)

        x = self.add_home_entry(self.dependency_manager.fsl, x)

        label_main_dep = QLabel(strings.mainwindow_home_label6)
        label_main_dep.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label_main_dep.setFont(bold_font)
        self.home_grid_layout.addWidget(label_main_dep, x, 0, 1, 2)
        x += 1

        x = self.add_home_entry(self.dependency_manager.freesurfer, x)
        self.global_config.freesurfer = self.dependency_manager.is_freesurfer()

        if DependencyManager.need_slicer_check(self.global_config):
            self.slicer_x = x
            x = self.add_home_entry(
                Dependence(
                    DependenceStatus.CHECKING, strings.mainwindow_dep_slicer_src
                ),
                x,
            )
            DependencyManager.check_slicer(
                self.global_config.get_slicer_path(), self.slicer_row
            )
        else:
            label = (
                strings.check_dep_slicer_found % self.global_config.get_slicer_version()
            )
            x = self.add_home_entry(Dependence(DependenceStatus.DETECTED, label), x)

        label_main_dep = QLabel(strings.mainwindow_home_label7)
        label_main_dep.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label_main_dep.setFont(bold_font)
        self.home_grid_layout.addWidget(label_main_dep, x, 0, 1, 2)
        x += 1

        x = self.add_home_entry(self.dependency_manager.graphviz, x)

        vertical_spacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )
        self.home_grid_layout.addItem(vertical_spacer, x, 0, 1, 2)

        self.homeTab.setLayout(self.home_grid_layout)

    def add_home_entry(self, dep: Dependence, x: int) -> int:
        """
        Generates a dependency check label, adding it to an existing layout

        Parameters
        ----------
        dep : Dependence
            A Dependence object to be parsed.
        x : int
            The starting grid layout row index.

        Returns
        -------
        int
            The next grid layout row index.

        """

        label_icon = QLabel()
        label_icon.setScaledContents(True)

        if dep.state == DependenceStatus.DETECTED:
            label_icon.setPixmap(self.OK_ICON)
        elif dep.state == DependenceStatus.WARNING:
            label_icon.setPixmap(self.WARNING_ICON)
        elif dep.state == DependenceStatus.CHECKING:
            label_icon = QSvgWidget()
            label_icon.load(self.LOADING_MOVIE_FILE)
        else:
            label_icon.setPixmap(self.ERROR_ICON)

        label_icon.setFixedSize(25, 25)

        old_icon_layout = self.home_grid_layout.itemAtPosition(x, 0)
        if old_icon_layout is not None:
            old_icon_layout.widget().deleteLater()
            self.home_grid_layout.removeItem(old_icon_layout)
        self.home_grid_layout.addWidget(label_icon, x, 0)

        label = QLabel(dep.label)
        label.setOpenExternalLinks(True)
        label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        old_label_layout = self.home_grid_layout.itemAtPosition(x, 1)
        if old_label_layout is not None:
            old_label_layout.widget().deleteLater()
            self.home_grid_layout.removeItem(old_label_layout)
        self.home_grid_layout.addWidget(label, x, 1)

        return x + 1

    def slicer_row(
        self, slicer_path: str, slicer_version: str, msg: str, state: DependenceStatus
    ):
        """
        Generates the Slicer dependency check label and path, if 3D Slicer is found.

        Parameters
        ----------
        slicer_path : str
            The local 3D Slicer path.
        slicer_version: str
            The Slicer version found
        msg : str
            The label message.
        state: DependenceStatus
            A state from DependenceStatus.

        Returns
        -------
        None.

        """

        self.add_home_entry(Dependence(state, msg), self.slicer_x)

        if state is DependenceStatus.DETECTED:
            self.global_config.set_slicer_path(slicer_path)
            self.global_config.set_slicer_version(slicer_version)
            self.global_config.save()

        for tab in self.subject_tab_array:
            tab.export_results_button_update_state()
            tab.load_scene_button_update_state()



from nipype.interfaces.freesurfer import SampleToSurface
from nipype.interfaces.fsl import (
    FLIRT,
    IsotropicSmooth,
    ApplyWarp,
    ApplyMask,
    SwapDimensions,
    ApplyXFM,
    ImageMaths,
)
from nipype.pipeline.engine import Node
from swane.nipype_pipeline.workflows.tractography_workflow import SIDES
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.CustomDcm2niix import CustomDcm2niix
from swane.nipype_pipeline.nodes.ForceOrient import ForceOrient
from swane.nipype_pipeline.nodes.AsymmetryIndex import AsymmetryIndex
from swane.nipype_pipeline.nodes.Zscore import Zscore
from nipype.interfaces.utility import IdentityInterface
from configparser import SectionProxy
import swane_supplement
from swane.config.config_enums import BETWEEN_MOD_FLIRT_COST


def func_map_workflow(
    name: str,
    dicom_dir: str,
    is_freesurfer: bool,
    config: SectionProxy,
    base_dir: str = "/",
) -> CustomWorkflow:
    """
    Analysis for PET or ASL:
        - registration to reference;
        - z-score and asymmetry index maps;
        - projection on FreeSurfer pial surface.

    Parameters
    ----------
    name : str
        The workflow name.
    dicom_dir : path
        The file path of the DICOM files.
    is_freesurfer : bool
        True if the reconall is available.
    config: SectionProxy
        workflow settings.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    reference : path
        The reference image for the registration.
    brain_mask : path
        The brain mask from T13D bet command.
    freesurfer_subjects_dir : path
        The directory from FreeSurfer analysis.
    freesurfer_subject_id :
        The subject id from FreeSurfer analysis.
    bgROI : path
        Basal ganglia and thalami ROI.
    ref_2_sym_warp : path
        Nonlinear registration warp from T13D to symmetric atlas.
    ref_2_sym_invwarp : path
        Nonlinear registration inverse warp from symmetric atlas to T13D.

    Returns
    -------
    workflow : CustomWorkflow
        The func map workflow.

    Output Node Fields
    ----------
    registered_file : string
        Functional map in T13D reference space.
    surf_lh : path
        If FreeSurfer is available, functional map projection on LH pial surface.
    surf_rh : path
        If FreeSurfer is available, functional map projection on RH pial surface.
    zscore : path
        If FreeSurfer is available, internal z-score statistics compared to basal ganglia.
    zscore_surf_lh : list of strings
        If FreeSurfer is available, z-score projection on LH pial surface.
    zscore_surf_rh : list of strings
        If FreeSurfer is available, z-score projection on RH pial surface.
    ai : path
        If AI is enabled, asymmetry index map.
    ai_surf_lh : path
        If FreeSurfer is available, asymmetry index projection on LH pial surface.
    ai_surf_rh : path
        If FreeSurfer is available, asymmetry index projection on RH pial surface.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(
        IdentityInterface(
            fields=[
                "reference",
                "brain_mask",
                "freesurfer_subjects_dir",
                "freesurfer_subject_id",
                "bgROI",
                "ref_2_sym_warp",
                "ref_2_sym_invwarp",
            ]
        ),
        name="inputnode",
    )

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=[
                "registered_file",
                "surf_lh",
                "surf_rh",
                "zscore",
                "zscore_surf_lh",
                "zscore_surf_rh",
                "ai",
                "ai_surf_lh",
                "ai_surf_rh",
            ]
        ),
        name="outputnode",
    )

    # NODE 1: Conversion dicom -> nifti
    conversion = Node(CustomDcm2niix(), name="%s_conv" % name)
    conversion.inputs.out_filename = name
    conversion.inputs.bids_format = False
    conversion.inputs.source_dir = dicom_dir
    conversion.inputs.name_conflicts = 1
    conversion.inputs.merge_imgs = 2

    # NODE 2: Orienting in radiological convention
    reorient = Node(ForceOrient(), name="%s_reOrient" % name)
    workflow.connect(conversion, "converted_files", reorient, "in_file")

    # NODE 3: Gaussian smoothing
    smooth = Node(IsotropicSmooth(), name="%s_smooth" % name)
    smooth.long_name = "smoothing"
    smooth.inputs.sigma = 2
    workflow.connect(reorient, "out_file", smooth, "in_file")

    # NODE 4: Registration matrix calculation in reference space
    func_2_ref_flirt = Node(FLIRT(), name="%s_2_ref_flirt" % name)
    func_2_ref_flirt.long_name = "%s to reference space"
    if config.getenum_safe("cost_func") is BETWEEN_MOD_FLIRT_COST.MULTUAL_INFORMATION:
        cost = "mutualinfo"
    elif (
        config.getenum_safe("cost_func")
        is BETWEEN_MOD_FLIRT_COST.NORMALIZED_MUTUAL_INFORMATION
    ):
        cost = "normmi"
    else:
        cost = "corratio"
    func_2_ref_flirt.inputs.cost = cost
    func_2_ref_flirt.inputs.searchr_x = [-90, 90]
    func_2_ref_flirt.inputs.searchr_y = [-90, 90]
    func_2_ref_flirt.inputs.searchr_z = [-90, 90]
    func_2_ref_flirt.inputs.dof = 6
    func_2_ref_flirt.inputs.interp = "trilinear"
    workflow.connect(reorient, "out_file", func_2_ref_flirt, "in_file")
    workflow.connect(inputnode, "reference", func_2_ref_flirt, "reference")

    # NODE 5: Smooth volume linear transformation in reference space
    smooth_2_ref_flirt = Node(ApplyXFM(), name="%s_smooth_2_ref_flirt" % name)
    smooth_2_ref_flirt.long_name = "%s to reference space"
    smooth_2_ref_flirt.inputs.out_file = "r-%s.nii.gz" % name
    smooth_2_ref_flirt.inputs.interp = "trilinear"
    workflow.connect(smooth, "out_file", smooth_2_ref_flirt, "in_file")
    workflow.connect(inputnode, "reference", smooth_2_ref_flirt, "reference")
    workflow.connect(
        func_2_ref_flirt, "out_matrix_file", smooth_2_ref_flirt, "in_matrix_file"
    )

    # NODE 6: Scalp removal
    mask = Node(ApplyMask(), name="%s_mask" % name)
    mask.long_name = name + " %s"
    mask.inputs.out_file = "r-%s.nii.gz" % name
    workflow.connect(smooth_2_ref_flirt, "out_file", mask, "in_file")
    workflow.connect(inputnode, "brain_mask", mask, "mask_file")

    workflow.connect(mask, "out_file", outputnode, "registered_file")

    if is_freesurfer:
        # NODE 7: Projection of the map on FreeSurfer pial surface
        for side in SIDES:
            func_surf = Node(SampleToSurface(), name="%s_surf_%s" % (name, side))
            func_surf.long_name = side + " " + name + " %s"
            func_surf.inputs.hemi = side
            func_surf.inputs.out_file = "%s_surf_%s.mgz" % (name, side)
            func_surf.inputs.cortex_mask = True
            func_surf.inputs.reg_header = True
            func_surf.inputs.sampling_method = "point"
            func_surf.inputs.sampling_range = 0.5
            func_surf.inputs.sampling_units = "frac"
            workflow.connect(mask, "out_file", func_surf, "source_file")
            workflow.connect(
                inputnode, "freesurfer_subjects_dir", func_surf, "subjects_dir"
            )
            workflow.connect(
                inputnode, "freesurfer_subject_id", func_surf, "subject_id"
            )

            workflow.connect(func_surf, "out_file", outputnode, "surf_%s" % side)

        # NODE 8: z-score calculation
        zscore = Node(Zscore(), name="%s_zscore" % name)
        zscore.long_name = "internal zscore"
        zscore.inputs.out_file = "r-%s_zscore.nii.gz" % name
        workflow.connect(mask, "out_file", zscore, "in_file")
        workflow.connect(inputnode, "bgROI", zscore, "ROI_file")

        workflow.connect(zscore, "out_file", outputnode, "zscore")

        # NODE 10: Projection of z-score on FreeSurfer pial surface
        for side in SIDES:
            zscore_surf_lh = Node(
                SampleToSurface(), name="%s_zscore_surf_%s" % (name, side)
            )
            zscore_surf_lh.long_name = side + " zscore %s"
            zscore_surf_lh.inputs.hemi = side
            zscore_surf_lh.inputs.out_file = "%s_zscore_surf_%s.mgz" % (name, side)
            zscore_surf_lh.inputs.cortex_mask = True
            zscore_surf_lh.inputs.reg_header = True
            zscore_surf_lh.inputs.sampling_method = "point"
            zscore_surf_lh.inputs.sampling_range = 0.5
            zscore_surf_lh.inputs.sampling_units = "frac"
            workflow.connect(zscore, "out_file", zscore_surf_lh, "source_file")
            workflow.connect(
                inputnode, "freesurfer_subjects_dir", zscore_surf_lh, "subjects_dir"
            )
            workflow.connect(
                inputnode, "freesurfer_subject_id", zscore_surf_lh, "subject_id"
            )

            workflow.connect(
                zscore_surf_lh, "out_file", outputnode, "zscore_surf_%s" % side
            )

    is_ai = config.getboolean_safe("ai")

    if is_ai:
        sym_template = swane_supplement.sym_template

        # NODE 11: Nonlinear transformation of the images in symmetric atlas
        func_2_sym_warp = Node(ApplyWarp(), name="%s_2_sym_warp" % name)
        func_2_sym_warp.long_name = "%s to symmetric atlas"
        func_2_sym_warp.inputs.ref_file = sym_template
        workflow.connect(smooth_2_ref_flirt, "out_file", func_2_sym_warp, "in_file")
        workflow.connect(inputnode, "ref_2_sym_warp", func_2_sym_warp, "field_file")

        # NODE 12: RL swap of image in symmetric atlas
        sym_swap = Node(SwapDimensions(), name="%s_sym_swap" % name)
        sym_swap.long_name = "right-left flip"
        sym_swap.inputs.out_file = "%s_sym_swapped.nii.gz" % name
        sym_swap.inputs.new_dims = ("-x", "y", "z")
        workflow.connect(func_2_sym_warp, "out_file", sym_swap, "in_file")

        # NODE 13: Asymmetry index calculation
        ai = Node(AsymmetryIndex(), name="%s_ai" % name)
        ai.long_name = "asymmetry index"
        ai.inputs.out_file = "r-%s_ai.nii.gz" % name
        workflow.connect(func_2_sym_warp, "out_file", ai, "in_file")
        workflow.connect(sym_swap, "out_file", ai, "swapped_file")

        # NODE 14: AI thresholding
        ai_threshold = Node(ImageMaths(), name="%s_ai_threshold" % name)
        threshold = config.getint_safe("ai_threshold")
        threshold = abs(threshold / 100)
        ai_threshold.inputs.op_string = "-thr %f -uthr %f" % (-threshold, threshold)
        workflow.connect(ai, "out_file", ai_threshold, "in_file")

        # NODE 15: AI Nonlinear transformation to reference space
        ai_2_ref = Node(ApplyWarp(), name="%s_ai_2_ref" % name)
        ai_2_ref.long_name = "asymmetry index %s from symmetric atlas"
        ai_2_ref.inputs.out_file = "r-%s_ai.nii.gz" % name
        workflow.connect(ai_threshold, "out_file", ai_2_ref, "in_file")
        workflow.connect(inputnode, "ref_2_sym_invwarp", ai_2_ref, "field_file")
        workflow.connect(inputnode, "reference", ai_2_ref, "ref_file")

        # NODE 16: AI scalp removal
        ai_mask = Node(ApplyMask(), name="%s_ai_mask" % name)
        ai_mask.long_name = name + " AI %s"
        ai_mask.inputs.out_file = "r-%s_ai.nii.gz" % name
        workflow.connect(ai_2_ref, "out_file", ai_mask, "in_file")
        workflow.connect(inputnode, "brain_mask", ai_mask, "mask_file")

        workflow.connect(ai_mask, "out_file", outputnode, "ai")

        if is_freesurfer:
            for side in SIDES:
                # NODE 17: Projection of AI on FreeSurfer pial surface
                ai_surf = Node(SampleToSurface(), name="%s_ai_surf_%s" % (name, side))
                ai_surf.long_name = side + " asymmetry index %s"
                ai_surf.inputs.hemi = side
                ai_surf.inputs.out_file = "%s_ai_surf_%s.mgz" % (name, side)
                ai_surf.inputs.cortex_mask = True
                ai_surf.inputs.reg_header = True
                ai_surf.inputs.sampling_method = "point"
                ai_surf.inputs.sampling_range = 0.5
                ai_surf.inputs.sampling_units = "frac"
                workflow.connect(ai_mask, "out_file", ai_surf, "source_file")
                workflow.connect(
                    inputnode, "freesurfer_subjects_dir", ai_surf, "subjects_dir"
                )
                workflow.connect(
                    inputnode, "freesurfer_subject_id", ai_surf, "subject_id"
                )

                workflow.connect(ai_surf, "out_file", outputnode, "ai_surf_%s" % side)

    return workflow



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl import UnaryMaths, Threshold
from os.path import abspath
import os
from nipype.interfaces.base import (
    traits,
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
    isdefined,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class ThrROIInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input image")
    seg_val_min = traits.Float(
        mandatory=True, desc="the min value of interested segmentation"
    )
    seg_val_max = traits.Float(
        mandatory=True, desc="the max value of interested segmentation"
    )
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class ThrROIOutputSpec(TraitedSpec):
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class ThrROI(BaseInterface):
    """
    Extracts a binary ROI from a segmentation using a min and a max value.

    """

    input_spec = ThrROIInputSpec
    output_spec = ThrROIOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()

        below = Threshold()
        below.inputs.in_file = self.inputs.in_file
        below.inputs.direction = "below"
        below.inputs.thresh = self.inputs.seg_val_min
        below.inputs.out_file = abspath(
            "below_" + os.path.basename(self.inputs.in_file)
        )
        below_res = below.run()

        above = Threshold()
        above.inputs.in_file = below_res.outputs.out_file
        above.inputs.direction = "above"
        above.inputs.thresh = self.inputs.seg_val_max
        above.inputs.out_file = abspath(
            "above_" + os.path.basename(self.inputs.in_file)
        )
        above_res = above.run()

        bin = UnaryMaths()
        bin.inputs.in_file = above_res.outputs.out_file
        bin.inputs.operation = "bin"
        bin.inputs.out_file = self.inputs.out_file
        bin.run()

        # todo semplificare il nodo subclassando ImageMaths, esempio sotto
        # calc_roi = ImageMaths()
        # calc_roi.inputs.in_file = self.inputs.in_file
        # calc_roi.inputs.op_string = "-thr %f -uthr %f -bin" % (self.inputs.seg_val_min, self.inputs.seg_val_max)
        # calc_roi.inputs.out_file = self.inputs.out_file
        # calc_roi.run()

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = (
                "ROI_"
                + str(self.inputs.seg_val_min)
                + "_"
                + str(self.inputs.seg_val_min)
                + "_"
                + os.path.basename(self.inputs.in_file)
            )
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl import ProbTrackX2
from nipype.interfaces.fsl.dti import ProbTrackX2InputSpec
from nipype.interfaces.base import traits, isdefined


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.dti.ProbTrackX2InputSpec)  -*-
class CustomProbTrackX2InputSpec(ProbTrackX2InputSpec):
    use_gpu = traits.Bool(False, desc="Use the GPU version of probtrackx")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.ProbTrackX2)  -*-
class CustomProbTrackX2(ProbTrackX2):
    """
    Custom implementation of ProbTrackX2 Nipype Node to support --rseed as Int and --sampvox as Float.

    """

    input_spec = CustomProbTrackX2InputSpec
    _default_cmd = ProbTrackX2._cmd

    def __init__(self, **inputs):
        super().__init__(**inputs)
        self.inputs.on_trait_change(self._cuda_update, "use_gpu")

    def _cuda_update(self):
        if isdefined(self.inputs.use_gpu) and self.inputs.use_gpu:
            self._cmd = "probtrackx2_gpu"
        else:
            self._cmd = self._default_cmd



from enum import Enum, auto

from swane.config.PrefCategory import PrefCategory


class InputTypes(Enum):
    TEXT = auto()
    PASSWORD = auto()
    INT = auto()
    BOOLEAN = auto()
    ENUM = auto()
    FILE = auto()
    DIRECTORY = auto()
    FLOAT = auto()


class WORKFLOW_TYPES(Enum):
    STRUCTURAL = "Structural Workflow"
    FUNCTIONAL = "Morpho-Functional Workflow"


class SLICER_EXTENSIONS(Enum):
    MRB = "mrb"
    MRML = "mrml"


class CORE_LIMIT(Enum):
    NO_LIMIT = "No limit"
    SOFT_CAP = "Soft cap"
    HARD_CAP = "Hard Cap"


class BETWEEN_MOD_FLIRT_COST(Enum):
    MULTUAL_INFORMATION = "Mutual information"
    NORMALIZED_MUTUAL_INFORMATION = "Normalized mutual information"
    CORRELATION_RATIO = "Correlation ratio"


class VEIN_DETECTION_MODE(Enum):
    SD = "Automatic (standard deviation)"
    MEAN = "Automatic (mean value)"
    FIRST = "Always first volume"
    SECOND = "Always second volume"


class BLOCK_DESIGN(Enum):
    RARA = "rArA..."
    RARB = "rArBrArB..."


class SLICE_TIMING(Enum):
    UNKNOWN = "Unknown"
    UP = "Regular up"
    DOWN = "Regular down"
    INTERLEAVED = "Interleaved"


class ImageModality(Enum):
    RM = "mr"
    PET = "pt"

    @staticmethod
    def from_string(mod_string: str):
        for mod in ImageModality:
            if mod.value.lower() == mod_string.lower():
                return mod
        return None


class PLANES(Enum):
    TRA = "transverse"
    COR = "coronal"
    SAG = "sagittal"


class GlobalPrefCategoryList(Enum):
    MAIN = PrefCategory("main", "Global settings")
    PERFORMANCE = PrefCategory("performance", "Performance")
    OPTIONAL_SERIES = PrefCategory("optional_series", "Optional series")
    MAIL_SETTINGS = PrefCategory("mail_settings", "Mail settings")

    def __str__(self):
        return self.value.name



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

import shutil
from nipype.interfaces.fsl import ExtractROI
from nipype import Node
from os.path import abspath
import os
from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterface,
    BaseInterfaceInputSpec,
    File,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class DeleteVolumesInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input image")
    nvols = traits.Int(mandatory=True, desc="original file volumes")
    del_start_vols = traits.Int(mandatory=True, desc="volumes to delete from start")
    del_end_vols = traits.Int(mandatory=True, desc="volumes to delete from end")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class DeleteVolumesOutputSpec(TraitedSpec):
    out_file = File(desc="the output image")
    nvols = traits.Int(desc="new number of volumes")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class DeleteVolumes(BaseInterface):
    """
    Removes specified num. of volumes from start and end of a 4d NIFTI file.

    """

    input_spec = DeleteVolumesInputSpec
    output_spec = DeleteVolumesOutputSpec

    def _run_interface(self, runtime):
        out_file = self._gen_outfilename()
        if self.inputs.del_start_vols == 0 and self.inputs.del_end_vols == 0:
            shutil.copy(self.inputs.in_file, out_file)
        else:
            fmri_delete_volumes = Node(ExtractROI(), name="fMRI_delete_volumes")
            fmri_delete_volumes.inputs.in_file = self.inputs.in_file
            fmri_delete_volumes.inputs.t_min = self.inputs.del_start_vols
            fmri_delete_volumes.inputs.t_size = (
                self.inputs.nvols
                - self.inputs.del_start_vols
                - self.inputs.del_end_vols
            )
            fmri_delete_volumes.inputs.roi_file = out_file

            fmri_delete_volumes.run()

        return runtime

    def _gen_outfilename(self):
        out_file = os.path.basename(self.inputs.in_file)
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        outputs["nvols"] = (
            self.inputs.nvols - self.inputs.del_start_vols - self.inputs.del_end_vols
        )
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterface,
    BaseInterfaceInputSpec,
)
from scipy.stats import ttest_ind_from_stats


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class TTestInputSpec(BaseInterfaceInputSpec):
    stats_lh = traits.List(mandatory=True, desc="Stats for left side")
    stats_rh = traits.List(mandatory=True, desc="Stats for right side")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class TTestOutputSpec(TraitedSpec):
    stat_t = traits.Float(desc="T statistics")
    stat_p = traits.Float(desc="P value")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class TTest(BaseInterface):
    """
    Calculate T statistics of two given distributions

    """

    input_spec = TTestInputSpec
    output_spec = TTestOutputSpec

    def _run_interface(self, runtime):
        try:
            self.t, self.p = ttest_ind_from_stats(
                mean1=self.inputs.stats_lh[0],
                std1=self.inputs.stats_lh[1],
                nobs1=self.inputs.stats_lh[2],
                mean2=self.inputs.stats_rh[0],
                std2=self.inputs.stats_rh[1],
                nobs2=self.inputs.stats_rh[2],
            )
        except:
            self.t = 0
            self.p = 0

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["stat_t"] = self.t
        outputs["stat_p"] = self.p
        return outputs



from datetime import datetime
from functools import partial
import os
from PySide6.QtWidgets import (
    QDialog,
    QGridLayout,
    QVBoxLayout,
    QWidget,
    QPushButton,
    QSpacerItem,
    QSizePolicy,
    QMessageBox,
)
from swane import strings, EXIT_CODE_REBOOT
from swane.ui.PreferenceUIEntry import PreferenceUIEntry
from swane.config.preference_list import WF_PREFERENCES, GLOBAL_PREFERENCES
from PySide6_VerticalQTabWidget import VerticalQTabWidget
from swane.config.config_enums import InputTypes, GlobalPrefCategoryList
from swane.utils.DataInputList import DataInputList
from enum import Enum
from swane.utils.CryptographyManager import CryptographyManager

from swane.utils.MailManager import MailManager


class PreferencesWindow(QDialog):
    """
    Custom implementation of PySide QDialog to show SWANe workflow preferences.

    """

    def __init__(self, my_config, dependency_manager, is_workflow: bool, parent=None):
        super(PreferencesWindow, self).__init__(parent)

        self.my_config = my_config
        self.dependency_manager = dependency_manager
        self.restart = False

        if self.my_config.global_config:
            if is_workflow:
                title = strings.menu_wf_pref
                self.preferences = WF_PREFERENCES
            else:
                title = strings.menu_pref
                self.preferences = GLOBAL_PREFERENCES
        else:
            self.preferences = WF_PREFERENCES
            title = (
                os.path.basename(os.path.dirname(self.my_config.config_file))
                + strings.wf_pref_window_title_user
            )

        if is_workflow:
            default_pref_list = DataInputList
        else:
            default_pref_list = GlobalPrefCategoryList

        self.setWindowTitle(title)

        self.inputs = {}
        self.input_keys = {}

        layout = QVBoxLayout()

        tab_widget = VerticalQTabWidget(force_top_valign=True)

        x = 0

        # Define here to prevent calling during invalid pref value reading
        self.saveButton = QPushButton(strings.pref_window_save_button)
        self.saveButton.clicked.connect(self.save_preferences)

        for category in default_pref_list:
            if str(category) not in my_config:
                continue
            if (
                is_workflow
                and not my_config.global_config
                and not self.parent().subject.input_state_list[category].loaded
            ):
                continue

            cat_label = category.value.label

            self.input_keys[category] = {}

            tab = QWidget()
            tab_widget.addTab(tab, cat_label)
            grid = QGridLayout()
            tab.setLayout(grid)

            for key in my_config[category].keys():
                if key not in self.preferences[category]:
                    continue
                if self.preferences[category][key].hidden:
                    continue
                self.input_keys[category][key] = x
                self.inputs[x] = PreferenceUIEntry(
                    category=category,
                    key=key,
                    my_config=my_config,
                    entry=self.preferences[category][key],
                    parent=self,
                )
                # External dependencies check
                self.check_dependency(category, key, x)

                # Other preferences requirements
                if self.preferences[category][key].pref_requirement is not None:
                    # Search all inputs that are preference for this one and apply them a function on change
                    for pref_cat in self.preferences[category][key].pref_requirement:
                        if str(pref_cat) not in my_config:
                            continue
                        for pref_req in self.preferences[category][
                            key
                        ].pref_requirement[pref_cat]:
                            if str(pref_req[0]) not in my_config[pref_cat]:
                                continue
                            target_x = self.input_keys[pref_cat][pref_req[0]]
                            self.inputs[target_x].connect_change(
                                lambda checked, my_cat=category, my_key=key: self.requirement_changed(
                                    checked, my_cat, my_key
                                )
                            )
                    # After the loop check if pref should be enabled
                    self.requirement_changed(False, category, key)

                # Data input requirements
                if (
                    not my_config.global_config
                    and self.preferences[category][key].input_requirement is not None
                ):
                    for input_req in self.preferences[category][key].input_requirement:
                        for cat_check in default_pref_list:
                            if (
                                cat_check == input_req
                                and not self.parent()
                                .subject.input_state_list[cat_check]
                                .loaded
                            ):
                                self.inputs[x].disable(
                                    self.preferences[category][
                                        key
                                    ].input_requirement_fail_tooltip
                                )
                                break

                # Add GUI elements to grid
                grid.addWidget(self.inputs[x].label, x, 0)
                grid.addWidget(self.inputs[x].input_field, x, 1)
                if self.inputs[x].button is not None:
                    grid.addWidget(self.inputs[x].button, x, 2)
                x += 1

                # Informative text displayed in next row, if present
                if self.inputs[x - 1].informative_text_label is not None:
                    grid.addWidget(
                        self.inputs[x - 1].informative_text_label, x, 0, 1, 2
                    )
                    x += 1

            vertical_spacer = QSpacerItem(
                20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
            )
            grid.addItem(vertical_spacer, x, 0, 1, 2)

            if category is GlobalPrefCategoryList.MAIL_SETTINGS:
                x += 1
                test_button = QPushButton()
                test_button.setText(strings.pref_window_mail_test_button)
                test_button.setToolTip(strings.pref_window_mail_test_hint)

                test_button.clicked.connect(self.send_test_mail)

                grid.addWidget(test_button, x, 0, 1, 2)

        layout.addWidget(tab_widget)
        layout.addWidget(self.saveButton)

        discard_button = QPushButton(strings.pref_window_discard_button)
        discard_button.clicked.connect(self.close)
        layout.addWidget(discard_button)

        if self.preferences != GLOBAL_PREFERENCES:
            reset_button = QPushButton()
            if self.my_config.global_config:
                reset_button.setText(strings.pref_window_reset_global_button)
            else:
                reset_button.setText(strings.pref_window_reset_subj_button)
            reset_button.clicked.connect(self.reset)
            layout.addWidget(reset_button)

        self.setLayout(layout)

    def send_test_mail(self):
        try:
            server_address = self.inputs[
                self.input_keys[GlobalPrefCategoryList.MAIL_SETTINGS]["address"]
            ].get_value()
            port = self.inputs[
                self.input_keys[GlobalPrefCategoryList.MAIL_SETTINGS]["port"]
            ].get_value()
            username = self.inputs[
                self.input_keys[GlobalPrefCategoryList.MAIL_SETTINGS]["username"]
            ].get_value()
            password = CryptographyManager.decrypt(
                self.inputs[
                    self.input_keys[GlobalPrefCategoryList.MAIL_SETTINGS]["password"]
                ].get_value()
            )
            use_ssl = self.inputs[
                self.input_keys[GlobalPrefCategoryList.MAIL_SETTINGS]["use_ssl"]
            ].get_value()
            use_tls = self.inputs[
                self.input_keys[GlobalPrefCategoryList.MAIL_SETTINGS]["use_tls"]
            ].get_value()

            if server_address == "" or port == "" or username == "" or password == "":
                raise Exception("Fill all mail configuration fields first")

            mail_manager = MailManager(
                server_address, port, username, password, use_ssl, use_tls
            )

            mail_manager.send_report(
                f"This is a test mail sent by SWANe at {datetime.now()} to check the mail settings configuration inserted by the user"
            )
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Information)
            msg_box.setText(strings.pref_window_mail_test_success)
            msg_box.setWindowTitle("Mail report")
            msg_box.exec()
        except Exception as e:
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Critical)
            msg_box.setText(strings.pref_window_mail_test_fail)
            msg_box.setInformativeText(str(e))
            msg_box.setWindowTitle("Mail report")
            msg_box.exec()

    def check_dependency(self, category: Enum, key: str, x: int) -> bool:
        """
        Check an external dependence to test if a preference can be enabled.

        Parameters
        ----------
        category: Enum
            The category of the preference to be tested.
        key: str
            The name of the preference to be tested.
        x: int
            The index of the preference to be tested in the input field list.

        Return
        ----------
        True if external dependence is satisfied
        """
        if self.preferences[category][key].dependency is not None:
            dep_check = getattr(
                self.dependency_manager,
                self.preferences[category][key].dependency,
                None,
            )
            if dep_check is None or not callable(dep_check) or not dep_check():
                self.inputs[x].disable(
                    self.preferences[category][key].dependency_fail_tooltip
                )
                return False
        return True

    def requirement_changed(self, checked, my_cat: str, my_key: str):
        """
        Called if the user change a preference that is a requirement for another preference.
        Parameters
        ----------
        checked:
            Unused but passed by the event connection.
        my_cat: str
            The category of the preference to be tested.
        my_key: str
            The name of the preference to be tested.

        """
        my_x = self.input_keys[my_cat][my_key]
        if not self.check_dependency(my_cat, my_key, my_x):
            return
        pref_requirement = self.preferences[my_cat][my_key].pref_requirement
        for req_cat in pref_requirement:
            if req_cat not in self.input_keys:
                continue
            for req_key in pref_requirement[req_cat]:
                if req_key[0] not in self.input_keys[req_cat]:
                    continue
                req_x = self.input_keys[req_cat][req_key[0]]

                if self.inputs[req_x].input_type == InputTypes.BOOLEAN:
                    check = req_key[1] == self.inputs[req_x].input_field.isChecked()
                elif self.inputs[req_x].input_type == InputTypes.ENUM:
                    check = (
                        req_key[1].name
                        == self.inputs[req_x]
                        .input_field.itemData(
                            self.inputs[req_x].input_field.currentIndex()
                        )
                        .name
                    )
                else:
                    check = req_key[1] == self.inputs[req_x].input_field.get_value()

                if not check:
                    self.inputs[my_x].disable(
                        self.preferences[my_cat][my_key].pref_requirement_fail_tooltip
                    )
                    return
        self.inputs[my_x].enable()

    def save_preferences(self):
        """
        Loop all input fields and save values to configuration file.

        """
        for pref_entry in self.inputs.values():
            if pref_entry.changed:
                self.my_config[pref_entry.category][
                    pref_entry.key
                ] = pref_entry.get_value()

        self.my_config.save()
        if self.restart:
            ret_code = EXIT_CODE_REBOOT
        else:
            ret_code = 1

        self.done(ret_code)

    def reset(self):
        """
        Load default workflow settings and save them to the configuration file

        """
        msg_box = QMessageBox()
        if self.my_config.global_config:
            msg_box.setText(strings.pref_window_reset_global_box)
        else:
            msg_box.setText(strings.pref_window_reset_subj_box)
        msg_box.setIcon(QMessageBox.Icon.Warning)
        msg_box.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        msg_box.setDefaultButton(QMessageBox.StandardButton.No)
        ret2 = msg_box.exec()

        if ret2 == QMessageBox.StandardButton.Yes:
            self.done(-1)

    def set_restart(self):
        """
        Called when user change a settings that require SWANe restart.

        """
        self.restart = True
        self.saveButton.setText(strings.pref_window_save_restart_button)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

import shutil
from nipype.interfaces.fsl import SliceTimer
from nipype import Node
from os.path import abspath
import os
from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    BaseInterface,
    BaseInterfaceInputSpec,
    File,
)
from swane.config.config_enums import SLICE_TIMING


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class CustomSliceTimerInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input image")
    time_repetition = traits.Float(mandatory=True)
    slice_timing = traits.Enum(SLICE_TIMING, usedefault=True)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class CustomSliceTimerOutputSpec(TraitedSpec):
    slice_time_corrected_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class CustomSliceTimer(BaseInterface):
    """
    Applies a slice timing correction.

    """

    input_spec = CustomSliceTimerInputSpec
    output_spec = CustomSliceTimerOutputSpec

    def _run_interface(self, runtime):
        out_file = self._gen_outfilename()
        if self.inputs.slice_timing == SLICE_TIMING.UNKNOWN:
            shutil.copy(self.inputs.in_file, out_file)
        else:
            fmri_timing_correction = Node(SliceTimer(), name="fMRI_timing_correction")
            fmri_timing_correction.inputs.time_repetition = self.inputs.time_repetition
            fmri_timing_correction.inputs.in_file = self.inputs.in_file
            fmri_timing_correction.inputs.out_file = out_file
            if self.inputs.slice_timing == SLICE_TIMING.DOWN:
                fmri_timing_correction.inputs.index_dir = True
            elif self.inputs.slice_timing == SLICE_TIMING.INTERLEAVED:
                fmri_timing_correction.inputs.interleaved = True
            fmri_timing_correction.run()

        return runtime

    def _gen_outfilename(self):
        out_file = os.path.basename(self.inputs.in_file)
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["slice_time_corrected_file"] = self._gen_outfilename()
        return outputs



# slicerpython script for module checking
# Warning: slicer library is not required beacuse the script is executed in Slicer environment
import sys

extensionName = "SlicerFreeSurfer"
em = slicer.app.extensionsManagerModel()
em.interactive = False  # prevent display of popups
restart = False
if not em.installExtensionFromServer(extensionName, restart):
    raise ValueError(f"Failed to install {extensionName} extension")
sys.exit(0)



import os
from shutil import which
from nipype.interfaces import dcm2nii, fsl, freesurfer
from swane import strings
from packaging import version
from swane.config.ConfigManager import ConfigManager
from PySide6.QtCore import QThreadPool
from enum import Enum, auto
from nipype.utils.gpu_count import gpu_count
import platform


class DependenceStatus(Enum):
    DETECTED = auto()
    WARNING = auto()
    MISSING = auto()
    CHECKING = auto()


class Dependence:
    """
    An object with a dependence information
    """

    def __init__(
        self,
        state: DependenceStatus,
        label: str,
        state2: DependenceStatus = DependenceStatus.MISSING,
    ):
        """
        Parameters
        ----------
        state: DependenceStatus
            A value describing the dependence status
        label: str
            A string to inform the user about the depence status
        state2: DependenceStatus
            A value describing a subdependence status
        """

        self.state: DependenceStatus = state
        self.state2 = DependenceStatus.MISSING
        self.label = None
        self.update(state, label, state2)

    def update(
        self,
        state: DependenceStatus,
        label: str,
        state2: DependenceStatus = DependenceStatus.MISSING,
    ):
        """
        Parameters
        ----------
        state: DependenceStatus
            A value in DependenceStatus describing the dependence status
        label: str
            A string to inform the user about the depence status
        state2: DependenceStatus
            A value in DependenceStatus describing a subdependence status
        """

        if state in DependenceStatus:
            self.state = state
            self.label = label
            self.state2 = state2
        else:
            self.state = DependenceStatus.MISSING
            self.state2 = DependenceStatus.MISSING
            self.label = strings.check_dep_generic_error


class DependencyManager:
    """
    Manager for checking dependencies and reporting messages about them
    """

    MIN_FSL_VERSION = "6.0.6"
    MIN_FREESURFER_VERSION = "7.3.2"
    MIN_SLICER_VERSION = "5.2.1"
    FREESURFER_MATLAB_COMMAND = "checkMCR.sh"
    FSL_TCSH_COMMAND = "tcsh"
    FLS_LOCALE_COMMAND = "locale -a | grep en_US.utf8 >/dev/null || false "

    def __init__(self):
        self.dcm2niix = DependencyManager.check_dcm2niix()
        self.fsl = DependencyManager.check_fsl()
        self.freesurfer = DependencyManager.check_freesurfer()
        self.graphviz = DependencyManager.check_graphviz()

    def is_fsl(self) -> bool:
        """
        Returns
        -------
        True if fsl is detected (even if outdated).

        """
        return self.fsl.state != DependenceStatus.MISSING

    def is_dcm2niix(self) -> bool:
        """
        Returns
        -------
        True if dcm2niix is detected.

        """
        return self.dcm2niix.state == DependenceStatus.DETECTED

    def is_graphviz(self) -> bool:
        """
        Returns
        -------
        True if graphviz is detected.

        """
        return self.graphviz.state == DependenceStatus.DETECTED

    def is_freesurfer(self) -> bool:
        """
        Returns
        -------
        True if freesurfer is detected and configured (even if outdated).

        """
        return self.freesurfer.state != DependenceStatus.MISSING

    def is_freesurfer_matlab(self) -> bool:
        """
        Returns
        -------
        True if freesurfer matlab runtime is detected.

        """
        return self.freesurfer.state2 != DependenceStatus.MISSING

    @staticmethod
    def is_slicer(config: ConfigManager) -> bool:
        """
        Parameters
        ----------
        config: ConfigManager
            The global application preferences

        Returns
        -------
        True if the application preference has a valide Slicer path
        """

        if config is None or not config.global_config:
            return False

        current_slicer_path = config.get_slicer_path()
        if current_slicer_path == "" or not os.path.exists(current_slicer_path):
            return False

        return True

    @staticmethod
    def need_slicer_check(config: ConfigManager) -> bool:
        """
        Parameters
        ----------
        config: ConfigManager
            The global application preferences

        Returns
        -------
        check_slicer: bool
            True if the application needs to check for Slicer dependency
        """

        if config is None or not config.global_config:
            return False

        check_slicer = False
        if not DependencyManager.is_slicer(config):
            check_slicer = True

        if not DependencyManager.check_slicer_version(config.get_slicer_version()):
            check_slicer = True

        return check_slicer

    @staticmethod
    def check_slicer_version(slicer_version: str) -> bool:
        """

        Parameters
        ----------
        slicer_version: str
            The Slicer version to compare with MIN_SLICER_VERSION

        Returns
        -------
        True if the Slicer version is newer or equal to MIN_SLICER_VERSION

        """
        if slicer_version is None or slicer_version == "":
            return False
        try:
            return version.parse(slicer_version) >= version.parse(
                DependencyManager.MIN_SLICER_VERSION
            )
        except:
            return False

    @staticmethod
    def check_slicer(current_slicer_path: str, callback_func: callable):
        """
        Start a thread to scan the computer for Slicer
        Parameters
        ----------
        current_slicer_path: ConfigManager
            Current Slicer executable path
        callback_func: callable
            The UI function to call after the check thread
        """
        from swane.workers.SlicerCheckWorker import SlicerCheckWorker

        # if user set manually a Slicer path it is saved with a * as first character to force check that folder
        if current_slicer_path != "" and current_slicer_path[0] == "*":
            current_slicer_path = current_slicer_path[1:]
        if not os.path.exists(current_slicer_path):
            current_slicer_path = ""

        check_slicer_work = SlicerCheckWorker(current_slicer_path)
        check_slicer_work.signal.slicer.connect(callback_func)
        QThreadPool.globalInstance().start(check_slicer_work)

    @staticmethod
    def check_dcm2niix() -> Dependence:
        """
        Returns
        -------
        A Dependence object with dcm2niix information.
        """
        dcm2niix_version = dcm2nii.Info.version()
        if dcm2niix_version is None:
            return Dependence(
                DependenceStatus.MISSING, strings.check_dep_dcm2niix_error
            )
        return Dependence(
            DependenceStatus.DETECTED,
            strings.check_dep_dcm2niix_found % str(dcm2niix_version),
        )

    @staticmethod
    def check_fsl() -> Dependence:
        """
        Returns
        -------
        A Dependence object with fsl information.
        """
        fsl_version = fsl.base.Info.version()
        if fsl_version is None:
            return Dependence(DependenceStatus.MISSING, strings.check_dep_fsl_error)
        try:
            found_version = version.parse(fsl_version)
        except:
            found_version = version.parse("0")

        # check if locale en_US.utf8 is available
        if platform.system() == "Linux":
            locale_en = os.system(DependencyManager.FLS_LOCALE_COMMAND)
            if locale_en != 0:
                return Dependence(
                    DependenceStatus.MISSING, strings.check_dep_fsl_no_locale
                )

        # check fsl version
        if found_version < version.parse(DependencyManager.MIN_FSL_VERSION):
            return Dependence(
                DependenceStatus.WARNING,
                strings.check_dep_fsl_wrong_version
                % (fsl_version, DependencyManager.MIN_FSL_VERSION),
            )

        return Dependence(
            DependenceStatus.DETECTED, strings.check_dep_fsl_found % fsl_version
        )

    @staticmethod
    def check_graphviz() -> Dependence:
        """
        Returns
        -------
        A Dependence object with graphviz information.
        """
        if which("dot") is None:
            return Dependence(DependenceStatus.WARNING, strings.check_dep_graph_error)
        return Dependence(DependenceStatus.DETECTED, strings.check_dep_graph_found)

    @staticmethod
    def check_freesurfer() -> Dependence:
        """
        Returns
        -------
        A Dependence object with freesurfer and freesurfer matlab runtime information.
        """

        # FS installed
        if freesurfer.base.Info.version() is None:
            return Dependence(
                DependenceStatus.MISSING,
                strings.check_dep_fs_error1,
                DependenceStatus.MISSING,
            )
        freesurfer_version = str(freesurfer.base.Info.looseversion())

        # FS version file presence
        if "FREESURFER_HOME" not in os.environ:
            return Dependence(
                DependenceStatus.MISSING,
                strings.check_dep_fs_error2 % freesurfer_version,
                DependenceStatus.MISSING,
            )
        license_file = os.path.join(os.environ["FREESURFER_HOME"], "license.txt")
        if not os.path.exists(license_file):
            return Dependence(
                DependenceStatus.MISSING,
                strings.check_dep_fs_error4 % freesurfer_version,
                DependenceStatus.MISSING,
            )
        try:
            found_version = version.parse(freesurfer_version)
        except:
            found_version = version.parse("0")

        # FS minimum version
        if found_version < version.parse(DependencyManager.MIN_FREESURFER_VERSION):
            return Dependence(
                DependenceStatus.WARNING,
                strings.check_dep_fs_wrong_version
                % (freesurfer_version, DependencyManager.MIN_FREESURFER_VERSION),
            )

        # tcsh shell installed
        if which(DependencyManager.FSL_TCSH_COMMAND) is None:
            return Dependence(
                DependenceStatus.WARNING,
                strings.check_dep_fs_no_tcsh % freesurfer_version,
                DependenceStatus.MISSING,
            )

        # FS matlab runtime
        mrc = os.system(DependencyManager.FREESURFER_MATLAB_COMMAND)
        if mrc != 0:
            # TODO: facciamo un parse dell'output del comando per dare all'utente il comando di installazione? o forse è meglio non basarsi sul formato attuale dell'output e linkare direttamente la pagina ufficiale?
            return Dependence(
                DependenceStatus.WARNING,
                strings.check_dep_fs_error3 % freesurfer_version,
                DependenceStatus.MISSING,
            )
        return Dependence(
            DependenceStatus.DETECTED,
            strings.check_dep_fs_found % freesurfer_version,
            DependenceStatus.DETECTED,
        )

    @staticmethod
    def is_cuda():
        return gpu_count() > 0



import os
import subprocess
from PySide6.QtCore import QRunnable, Signal, QObject
from swane import strings
from swane.utils.DependencyManager import DependencyManager, DependenceStatus
import platform


class SlicerCheckSignaler(QObject):
    slicer = Signal(str, str, str, DependenceStatus)


class SlicerCheckWorker(QRunnable):
    """
    Spawn a thread for 3D Slicer dependency check

    """

    def __init__(self, current_slicer_path: str):
        super(SlicerCheckWorker, self).__init__()
        self.signal = SlicerCheckSignaler()
        self.current_slicer_path = current_slicer_path

    @staticmethod
    def find_slicer_python(current_slicer_path: str) -> (list[str], str):
        # If current_slicer_path doeas not exists, replace with a blank string
        # If it is a file, search in its directory
        if not os.path.exists(current_slicer_path):
            current_slicer_path = ""
        elif os.path.isfile(current_slicer_path):
            current_slicer_path = os.path.dirname(current_slicer_path)

        # Adjust search path based on OS
        if platform.system() == "Darwin":
            if current_slicer_path == "":
                src_path = "/Applications"
            else:
                src_path = current_slicer_path
            find_cmd = (
                "find "
                + src_path
                + " -type f -wholename *app/Contents/bin/PythonSlicer -print -quit 2>/dev/null"
            )
            rel_path = "../MacOS/Slicer"
        else:
            if current_slicer_path == "":
                src_path = "/"
            else:
                src_path = current_slicer_path
            find_cmd = (
                "find "
                + src_path
                + " -executable -type f -wholename *bin/PythonSlicer -print -quit 2>/dev/null"
            )
            rel_path = "../Slicer"

        # Perform search with find
        output = subprocess.run(
            find_cmd, shell=True, stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        split = output.split("\n")
        while "" in split:
            split.remove("")
        return split, rel_path

    def run(self):
        repeat = True
        cmd = ""
        state: DependenceStatus = DependenceStatus.MISSING
        label = ""
        slicer_version = ""

        while repeat:
            split, rel_path = SlicerCheckWorker.find_slicer_python(
                self.current_slicer_path
            )
            # find slicerpython executable and go back to slicer executable with rel_path
            for entry in split:
                cmd = os.path.abspath(os.path.join(os.path.dirname(entry), rel_path))
                break
            if cmd == "" or not os.path.exists(cmd):
                # if slicer executable is not found, search entire filesystem if we were searchng a specific folder
                # otherwise stop loop, slicer is not detectable on system
                if self.current_slicer_path != "":
                    self.current_slicer_path = ""
                else:
                    repeat = False
                label = strings.check_dep_slicer_error1
            else:
                # if slicer command is found, version check
                repeat = False
                cmd2 = cmd + " --version"
                output2 = subprocess.run(
                    cmd2, shell=True, stdout=subprocess.PIPE
                ).stdout.decode("utf-8")
                slicer_version = output2.replace("Slicer ", "").replace("\n", "")
                if not DependencyManager.check_slicer_version(slicer_version):
                    label = strings.check_dep_slicer_wrong_version % (
                        slicer_version,
                        DependencyManager.MIN_SLICER_VERSION,
                    )
                    state = DependenceStatus.WARNING
                else:
                    # if version check is passed, check for freesurfer module
                    # TODO: we can try to install missing feesurfer module automatically.
                    #   see slicer_script_freesurfer_module_install.py for example
                    cmd3 = (
                        cmd
                        + " --no-splash --no-main-window --python-script "
                        + os.path.join(
                            os.path.dirname(__file__),
                            "slicer_script_freesurfer_module_check.py",
                        )
                    )
                    output3 = subprocess.run(
                        cmd3, shell=True, stdout=subprocess.PIPE
                    ).stdout.decode("utf-8")
                    if "MODULE FOUND" in output3:
                        state = DependenceStatus.DETECTED
                        label = strings.check_dep_slicer_found % slicer_version
                    else:
                        label = strings.check_dep_slicer_error2

        self.signal.slicer.emit(cmd, slicer_version, label, state)

    def terminate(self):
        return



from nipype.interfaces.fsl import (
    BET,
    FLIRT,
    ConvertXFM,
    ExtractROI,
    EddyCorrect,
    DTIFit,
    ApplyXFM,
    FNIRT,
    BEDPOSTX5,
)
from nipype.pipeline.engine import Node
from swane.config.config_enums import CORE_LIMIT
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.CustomDcm2niix import CustomDcm2niix
from swane.nipype_pipeline.nodes.ForceOrient import ForceOrient
from swane.nipype_pipeline.nodes.GenEddyFiles import GenEddyFiles
from swane.nipype_pipeline.nodes.CustomEddy import CustomEddy
from configparser import SectionProxy
from nipype.interfaces.utility import IdentityInterface
from multiprocessing import cpu_count


def dti_preproc_workflow(
    name: str,
    dti_dir: str,
    config: SectionProxy,
    mni_dir: str = None,
    base_dir: str = "/",
    max_cpu: int = 0,
    multicore_node_limit: CORE_LIMIT = CORE_LIMIT.SOFT_CAP,
) -> CustomWorkflow:
    """
    DTI preprocessing workflow with eddy current and motion artifact correction.
    Diffusion metrics calculation and, if needed, bayesian estimation of
    diffusion parameters.

    Parameters
    ----------
    name : str
        The workflow name.
    dti_dir : path
        The directory path of DTI dicom files.
    mni_dir : path, optional
        The file path of the MNI template. The default is None.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".
    config: SectionProxy
        workflow settings.
    max_cpu : int, optional
        If greater than 0, limit the core usage of bedpostx. The default is 0.
    multicore_node_limit: CORE_LIMIT, optional
        Preference for bedpostX core usage. The default il CORE_LIMIT.SOFT_CAP

    Input Node Fields
    ----------
    ref_brain : path
        Betted T13D reference file.

    Returns
    -------
    workflow : CustomWorkflow
        The DTI preprocessing workflow.

    Output Node Fields
    ----------
    nodiff_mask_file : path
        Brain mask from b0.
    FA : path
        Fractional anysotropy map in reference space.
    fsamples : path
        Samples from the distribution of anysotropic volume fraction.
    phsamples : path
        Samples from the distribution on phi.
    thsamples : path
        Samples from the distribution on theta.
    diff2ref_mat : path
        Linear registration matrix from diffusion to T13D reference space.
    ref2diff_mat : path
        Linear registration inverse matrix from T13D reference to diffusion space.
    mni2ref_warp : path
        Nonlinear registration warp from MNI atlas to T13D reference space.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(IdentityInterface(fields=["ref_brain"]), name="inputnode")

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=[
                "nodiff_mask_file",
                "FA",
                "fsamples",
                "phsamples",
                "thsamples",
                "diff2ref_mat",
                "ref2diff_mat",
                "mni2ref_warp",
            ]
        ),
        name="outputnode",
    )

    is_cuda = config.getboolean_safe("cuda")

    # NODE 1: Conversion dicom -> nifti
    conversion = Node(CustomDcm2niix(), name="dti_conv")
    conversion.inputs.source_dir = dti_dir
    conversion.inputs.out_filename = "dti"
    conversion.inputs.bids_format = False
    conversion.inputs.request_dti = True
    conversion.inputs.name_conflicts = 1
    conversion.inputs.merge_imgs = 2

    # NODE 1b: Orienting in radiological convention
    reorient = Node(ForceOrient(), name="dti_reOrient")
    workflow.connect(conversion, "converted_files", reorient, "in_file")

    # NODE 2: b0 image extraction
    nodif = Node(ExtractROI(), name="dti_nodif")
    nodif.long_name = "b0 extraction"
    nodif.inputs.t_min = 0
    nodif.inputs.t_size = 1
    nodif.inputs.roi_file = "nodif.nii.gz"
    workflow.connect(reorient, "out_file", nodif, "in_file")

    # NODE 3: Scalp removal from b0 image
    bet = Node(BET(), name="nodif_BET")
    bet.inputs.frac = 0.3
    bet.inputs.robust = True
    bet.inputs.threshold = True
    bet.inputs.mask = True
    workflow.connect(nodif, "roi_file", bet, "in_file")

    old_eddy_correct = config.getboolean_safe("old_eddy_correct")
    if old_eddy_correct:
        # NODE 4a: Generate Eddy files
        eddy = Node(EddyCorrect(), name="dti_eddy")
        eddy.inputs.ref_num = 0
        eddy.inputs.out_file = "data.nii.gz"
        workflow.connect(reorient, "out_file", eddy, "in_file")
        eddy_output_name = "eddy_corrected"
    else:
        # NODE 4a: Generate Eddy files
        eddy_files = Node(GenEddyFiles(), name="dti_eddy_files")
        workflow.connect(conversion, "bvals", eddy_files, "bval")

        # NODE 4: Eddy current and motion artifact correction
        eddy = Node(CustomEddy(), name="dti_eddy")
        eddy.inputs.use_cuda = is_cuda
        if not is_cuda:
            if multicore_node_limit == CORE_LIMIT.HARD_CAP:
                eddy_cpu = max_cpu
                eddy.inputs.num_threads = max_cpu
            elif multicore_node_limit == CORE_LIMIT.SOFT_CAP:
                eddy_cpu = max_cpu
            else:
                eddy_cpu = cpu_count()
            eddy.inputs.environ = {
                "OMP_NUM_THREADS": str(eddy_cpu),
                "FSL_SKIP_GLOBAL": "1",
            }
            eddy.inputs.args = "--nthr=%d" % eddy_cpu

        workflow.connect(reorient, "out_file", eddy, "in_file")
        workflow.connect(conversion, "bvals", eddy, "in_bval")
        workflow.connect(conversion, "bvecs", eddy, "in_bvec")
        workflow.connect(eddy_files, "acqp", eddy, "in_acqp")
        workflow.connect(eddy_files, "index", eddy, "in_index")
        workflow.connect(bet, "mask_file", eddy, "in_mask")
        eddy_output_name = "out_corrected"

    # NODE 5: DTI metrics calculation
    dtifit = Node(DTIFit(), name="dti_dtifit")
    dtifit.long_name = "DTI metrics calculation"
    workflow.connect(eddy, eddy_output_name, dtifit, "dwi")
    workflow.connect(bet, "mask_file", dtifit, "mask")
    workflow.connect(conversion, "bvecs", dtifit, "bvecs")
    workflow.connect(conversion, "bvals", dtifit, "bvals")

    # NODE 6: b0 image linear registration in reference space
    flirt = Node(FLIRT(), name="diff2ref_FLIRT")
    flirt.long_name = "%s to reference space"
    flirt.inputs.out_matrix_file = "diff2ref.mat"
    flirt.inputs.cost = "corratio"
    flirt.inputs.searchr_x = [-90, 90]
    flirt.inputs.searchr_y = [-90, 90]
    flirt.inputs.searchr_z = [-90, 90]
    flirt.inputs.dof = 6
    workflow.connect(bet, "out_file", flirt, "in_file")
    workflow.connect(inputnode, "ref_brain", flirt, "reference")

    # NODE 7: FA linear transformation in reference space
    fa_2_ref_flirt = Node(ApplyXFM(), name="FA2ref_FLIRT")
    fa_2_ref_flirt.long_name = "FA %s in reference space"
    fa_2_ref_flirt.inputs.out_file = "r-FA.nii.gz"
    fa_2_ref_flirt.inputs.interp = "trilinear"
    workflow.connect(dtifit, "FA", fa_2_ref_flirt, "in_file")
    workflow.connect(flirt, "out_matrix_file", fa_2_ref_flirt, "in_matrix_file")
    workflow.connect(inputnode, "ref_brain", fa_2_ref_flirt, "reference")

    workflow.connect(fa_2_ref_flirt, "out_file", outputnode, "FA")

    is_tractography = config.getboolean_safe("tractography")

    if is_tractography:
        # NODE 1: Linear registration
        mni_2_ref_flirt = Node(FLIRT(), name="mni_2_ref_flirt")
        mni_2_ref_flirt.long_name = "atlas %s to diffusion space"
        mni_2_ref_flirt.inputs.searchr_x = [-90, 90]
        mni_2_ref_flirt.inputs.searchr_y = [-90, 90]
        mni_2_ref_flirt.inputs.searchr_z = [-90, 90]
        mni_2_ref_flirt.inputs.dof = 12
        mni_2_ref_flirt.inputs.cost = "corratio"
        mni_2_ref_flirt.inputs.out_matrix_file = "mni_2_ref.mat"
        mni_2_ref_flirt.inputs.in_file = mni_dir
        workflow.add_nodes([mni_2_ref_flirt])
        workflow.connect(inputnode, "ref_brain", mni_2_ref_flirt, "reference")

        # NODE 2: Nonlinear registration
        mni_2_ref_fnirt = Node(FNIRT(), name="mni_2_ref_fnirt")
        mni_2_ref_fnirt.long_name = "atlas %s to diffusion space"
        mni_2_ref_fnirt.inputs.fieldcoeff_file = True
        mni_2_ref_fnirt.inputs.in_file = mni_dir
        workflow.connect(
            mni_2_ref_flirt, "out_matrix_file", mni_2_ref_fnirt, "affine_file"
        )
        workflow.connect(inputnode, "ref_brain", mni_2_ref_fnirt, "ref_file")

        # NODE 8: Bayesian estimation of diffusion parameters
        bedpostx = Node(BEDPOSTX5(), name="dti_bedpostx")
        bedpostx.inputs.n_fibres = 2
        bedpostx.inputs.rician = True
        bedpostx.inputs.sample_every = 25
        bedpostx.inputs.n_jumps = 1250
        bedpostx.inputs.burn_in = 1000
        bedpostx.inputs.use_gpu = is_cuda
        if not is_cuda:
            # if cuda is enabled only 1 process is launched
            if multicore_node_limit == CORE_LIMIT.SOFT_CAP:
                bedpostx.inputs.environ = {"FSLSUB_PARALLEL": str(max_cpu)}
            elif multicore_node_limit == CORE_LIMIT.HARD_CAP:
                bedpostx.inputs.environ = {"FSLSUB_PARALLEL": str(max_cpu)}
                bedpostx.n_procs = max_cpu

        workflow.connect(eddy, eddy_output_name, bedpostx, "dwi")
        workflow.connect(bet, "mask_file", bedpostx, "mask")
        workflow.connect(conversion, "bvecs", bedpostx, "bvecs")
        workflow.connect(conversion, "bvals", bedpostx, "bvals")

        # NODE 9: Linear transformation inverse matrix calculation from diffusion to reference space
        ref2diff_convert = Node(ConvertXFM(), name="ref2diff_convert")
        ref2diff_convert.long_name = "inverse transformation from reference space"
        ref2diff_convert.inputs.invert_xfm = True
        ref2diff_convert.inputs.out_file = "ref2diff.mat"
        workflow.connect(flirt, "out_matrix_file", ref2diff_convert, "in_file")

        workflow.connect(bedpostx, "merged_fsamples", outputnode, "fsamples")
        workflow.connect(bet, "mask_file", outputnode, "nodiff_mask_file")
        workflow.connect(bedpostx, "merged_phsamples", outputnode, "phsamples")
        workflow.connect(bedpostx, "merged_thsamples", outputnode, "thsamples")

        workflow.connect(flirt, "out_matrix_file", outputnode, "diff2ref_mat")
        workflow.connect(ref2diff_convert, "out_file", outputnode, "ref2diff_mat")
        workflow.connect(mni_2_ref_fnirt, "fieldcoeff_file", outputnode, "mni2ref_warp")

    return workflow



import os
import glob
from nipype import Node, IdentityInterface, MapNode, JoinNode, Merge
from nipype.interfaces.fsl import ApplyWarp, ImageMaths
from configparser import SectionProxy
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.RandomSeedGenerator import RandomSeedGenerator
from swane.nipype_pipeline.nodes.CustomProbTrackX2 import CustomProbTrackX2
from swane.nipype_pipeline.nodes.MergeTargets import MergeTargets
from swane.nipype_pipeline.nodes.SumMultiTracks import SumMultiTracks
from swane.config.preference_list import TRACTS, DEFAULT_N_SAMPLES, XTRACT_DATA_DIR

SIDES = ["lh", "rh"]


def tractography_workflow(
    name: str, config: SectionProxy, base_dir: str = "/"
) -> CustomWorkflow:
    """
    Executes tractography for chosen tract using xtract protocols.

    Parameters
    ----------
    name : str
        The workflow and tract name.
    config: SectionProxy
        The subject workflow preferences.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    mask: path
        The ref brain mask.
    fsamples : path
        Samples from the distribution of anysotropic volume fraction.
    phsamples : path
        Samples from the distribution on phi.
    thsamples : path
        Samples from the distribution on theta.
    ref_brain : path
        Betted T13D.
    diff2ref_mat : path
        Linear registration matrix from diffusion to T13D reference space.
    ref2diff_mat : path
        Linear registration inverse matrix from T13D reference to diffusion space.
    mni2ref_warp : path
        Nonlinear registration warp from MNI atlas to T13D reference space.

    Returns
    -------
    workflow : CustomWorkflow
        The xtract workflow.

    Output Node Fields
    ----------
    fdt_paths_rh : path
        RH connectivity distribution in T13D reference space.
    fdt_paths_lh : path
        LH connectivity distribution in T13D reference space.
    waytotal_rh : path
        Text file containing a single number corresponding to the total number
        of generated tracts that have not been rejected by inclusion/exclusion
        mask criteria for RH side.
    waytotal_lh : path
        Text file containing a single number corresponding to the total number
        of generated tracts that have not been rejected by inclusion/exclusion
        mask criteria for LH side.

    """

    # Check if tract is in configuration list
    if name not in TRACTS:
        return None

    # Check for existance of xtract data directory and protocol name dicrectory
    if not os.path.exists(os.path.join(XTRACT_DATA_DIR, name + "_l")):
        return None

    workflow = CustomWorkflow(name="tract_" + name, base_dir=base_dir)

    inputnode = Node(
        IdentityInterface(
            fields=[
                "fsamples",
                "mask",
                "phsamples",
                "thsamples",
                "ref_brain",
                "diff2ref_mat",
                "ref2diff_mat",
                "mni2ref_warp",
            ]
        ),
        name="inputnode",
    )

    outputnode = Node(
        IdentityInterface(
            fields=["fdt_paths_rh", "fdt_paths_lh", "waytotal_rh", "waytotal_lh"]
        ),
        name="outputnode",
    )

    is_cuda = config.getboolean_safe("cuda")
    if is_cuda:
        # if cuda is enabled only 1 process is launched
        track_threads = 1
    else:
        track_threads = config.getint_safe("track_procs")

    # NODE 1: Random seed genration for cache preservation
    random_seed = Node(RandomSeedGenerator(), name="random_seed")
    random_seed.inputs.seeds_n = track_threads
    workflow.connect(inputnode, "mask", random_seed, "mask")

    try:
        n_samples = int(TRACTS[name][2] / track_threads)
    except:
        n_samples = int(DEFAULT_N_SAMPLES / track_threads)

    for side in SIDES:
        # Xtract protocol loading
        protocol_dir = os.path.join(XTRACT_DATA_DIR, name + "_" + side[0])

        seed_file = os.path.join(protocol_dir, "seed.nii.gz")
        exclude_file = os.path.join(protocol_dir, "exclude.nii.gz")
        stop_file = os.path.join(protocol_dir, "stop.nii.gz")
        target_files = glob.glob(os.path.join(protocol_dir, "target*"))

        invert_file = os.path.join(protocol_dir, "invert")
        wayorder_file = os.path.join(protocol_dir, "wayorder")

        is_invert = False
        is_wayorder = False

        if os.path.exists(invert_file):
            is_invert = True
        elif os.path.exists(wayorder_file):
            is_wayorder = True

        if not os.path.exists(seed_file):
            return None
        if len(target_files) == 0:
            return None

        # NODE 2: Seed ROI nonlinear transformation in T13D reference space
        seed_2_ref = Node(ApplyWarp(), name="seed_2_ref_%s_%s" % (name, side))
        seed_2_ref.long_name = side + " seed ROI %s"
        seed_2_ref.inputs.out_file = "r-seed_%s_%s.nii.gz" % (name, side)
        seed_2_ref.inputs.in_file = seed_file
        workflow.connect(inputnode, "ref_brain", seed_2_ref, "ref_file")
        workflow.connect(inputnode, "mni2ref_warp", seed_2_ref, "field_file")

        # NODE 3: Seed ROI bynarization
        seed_bin = Node(ImageMaths(), name="seed_bin_%s_%s" % (name, side))
        seed_bin.long_name = side + " seed ROI binarization"
        seed_bin.inputs.op_string = "-thr 0.1 -bin"
        seed_bin.inputs.out_data_type = "char"
        seed_bin.inputs.suffix = "_bin"
        workflow.connect(seed_2_ref, "out_file", seed_bin, "in_file")

        # NODE 4: Target ROIs nonlinear transformation in T13D reference space
        targets_2_ref = Node(ApplyWarp(), name="targets_2_ref_%s_%s" % (name, side))
        targets_2_ref.long_name = side + " target ROIs %s"
        if len(target_files) > 1:
            targets_2_ref.iterables = ("in_file", target_files)
        else:
            targets_2_ref.inputs.in_file = target_files[0]
        workflow.connect(inputnode, "ref_brain", targets_2_ref, "ref_file")
        workflow.connect(inputnode, "mni2ref_warp", targets_2_ref, "field_file")

        # NODE 5: Target ROIs bynarization
        targets_bin = Node(ImageMaths(), name="targets_bin_%s_%s" % (name, side))
        targets_bin.long_name = side + " target ROIs binarization"
        targets_bin.inputs.op_string = "-thr 0.1 -bin"
        targets_bin.inputs.out_data_type = "char"
        targets_bin.inputs.suffix = "_bin"
        workflow.connect(targets_2_ref, "out_file", targets_bin, "in_file")

        # NODE 10: Tractography
        probtrackx = MapNode(
            CustomProbTrackX2(),
            name="probtrackx_%s_%s" % (name, side),
            iterfield=["random_seed"],
        )
        probtrackx.long_name = side + " %s"
        probtrackx.inputs.n_samples = n_samples
        probtrackx.inputs.loop_check = True
        probtrackx.inputs.wayorder = is_wayorder
        probtrackx.inputs.rand_fib = 1
        probtrackx.inputs.sample_random_points = 1
        probtrackx.inputs.use_gpu = is_cuda
        # TODO argomento --ompl che fa??
        probtrackx.inputs.opd = True
        workflow.connect(inputnode, "fsamples", probtrackx, "fsamples")
        workflow.connect(inputnode, "mask", probtrackx, "mask")
        workflow.connect(inputnode, "ref_brain", probtrackx, "seed_ref")
        workflow.connect(inputnode, "phsamples", probtrackx, "phsamples")
        workflow.connect(inputnode, "thsamples", probtrackx, "thsamples")
        workflow.connect(inputnode, "ref2diff_mat", probtrackx, "xfm")
        workflow.connect(inputnode, "diff2ref_mat", probtrackx, "inv_xfm")
        workflow.connect(seed_bin, "out_file", probtrackx, "seed")
        workflow.connect(random_seed, "seeds", probtrackx, "random_seed")

        # Check the number of target ROIs
        if len(target_files) > 1:
            merge_targets = JoinNode(
                MergeTargets(),
                name="merge_targets_%s_%s" % (name, side),
                joinsource=targets_2_ref,
                joinfield="target_files",
            )
            merge_targets.long_name = side + " targets ROI merging"

            workflow.connect(targets_bin, "out_file", merge_targets, "target_files")

            workflow.connect(merge_targets, "out_file", probtrackx, "waypoints")
        else:
            workflow.connect(targets_bin, "out_file", probtrackx, "waypoints")

        # Check if inverted run is required in protocol
        if is_invert:
            # NODE 11: Inverted tractography
            probtrackx_inverted = MapNode(
                CustomProbTrackX2(),
                name="probtrackx_inverted_%s_%s" % (name, side),
                iterfield=["random_seed"],
            )
            probtrackx_inverted.long_name = side + " inverse %s"
            probtrackx_inverted.inputs.n_samples = n_samples
            probtrackx_inverted.inputs.loop_check = True
            probtrackx_inverted.inputs.wayorder = is_wayorder
            probtrackx_inverted.inputs.rand_fib = 1
            probtrackx_inverted.inputs.sample_random_points = 1
            probtrackx_inverted.inputs.opd = True
            probtrackx_inverted.inputs.use_gpu = is_cuda
            workflow.connect(inputnode, "fsamples", probtrackx_inverted, "fsamples")
            workflow.connect(inputnode, "mask", probtrackx_inverted, "mask")
            workflow.connect(inputnode, "ref_brain", probtrackx_inverted, "seed_ref")
            workflow.connect(inputnode, "phsamples", probtrackx_inverted, "phsamples")
            workflow.connect(inputnode, "thsamples", probtrackx_inverted, "thsamples")
            workflow.connect(inputnode, "ref2diff_mat", probtrackx_inverted, "xfm")
            workflow.connect(inputnode, "diff2ref_mat", probtrackx_inverted, "inv_xfm")
            workflow.connect(targets_bin, "out_file", probtrackx_inverted, "seed")
            workflow.connect(seed_bin, "out_file", probtrackx_inverted, "waypoints")
            workflow.connect(random_seed, "seeds", probtrackx_inverted, "random_seed")

        # Check for exclusion ROI in protocol
        if os.path.exists(exclude_file):
            # NODE 6: Exclusion ROI nonlinear transformation in T13D reference space
            exclude_2_ref = Node(ApplyWarp(), name="exclude_2_ref_%s_%s" % (name, side))
            exclude_2_ref.long_name = side + " exclusion ROI %s"
            exclude_2_ref.inputs.out_file = "r-exclude_%s_%s.nii.gz" % (name, side)
            exclude_2_ref.inputs.in_file = exclude_file
            workflow.connect(inputnode, "ref_brain", exclude_2_ref, "ref_file")
            workflow.connect(inputnode, "mni2ref_warp", exclude_2_ref, "field_file")

            # NODE 7: Exclusion ROI bynarization
            exclude_bin = Node(ImageMaths(), name="exclude_bin_%s_%s" % (name, side))
            exclude_bin.long_name = side + " exclusion ROI binarization"
            exclude_bin.inputs.op_string = "-thr 0.1 -bin"
            exclude_bin.inputs.out_data_type = "char"
            exclude_bin.inputs.suffix = "_bin"
            workflow.connect(exclude_2_ref, "out_file", exclude_bin, "in_file")

            workflow.connect(exclude_bin, "out_file", probtrackx, "avoid_mp")

            if is_invert:
                workflow.connect(
                    exclude_bin, "out_file", probtrackx_inverted, "avoid_mp"
                )

        # Check for stop ROI in protocol
        if os.path.exists(stop_file):
            # NODE 8: stop ROI nonlinear transformation in T13D reference space
            stop_2_ref = Node(ApplyWarp(), name="stop_2_ref_%s_%s" % (name, side))
            stop_2_ref.long_name = side + " stop ROI %s"
            stop_2_ref.inputs.out_file = "r-stop_%s_%s.nii.gz" % (name, side)
            stop_2_ref.inputs.in_file = stop_file
            workflow.connect(inputnode, "ref_brain", stop_2_ref, "ref_file")
            workflow.connect(inputnode, "mni2ref_warp", stop_2_ref, "field_file")

            # NODE 9: stop ROI bynarization
            stop_bin = Node(ImageMaths(), name="stop_bin_%s_%s" % (name, side))
            stop_bin.long_name = side + " stop ROI binarization"
            stop_bin.inputs.op_string = "-thr 0.1 -bin"
            stop_bin.inputs.out_data_type = "char"
            stop_bin.inputs.suffix = "_bin"
            workflow.connect(stop_2_ref, "out_file", stop_bin, "in_file")

            workflow.connect(stop_bin, "out_file", probtrackx, "stop_mask")

            if is_invert:
                workflow.connect(stop_bin, "out_file", probtrackx_inverted, "stop_mask")

        # NODE 14: Sum tractography and inverted tractography results
        sum_multi_tracks = Node(SumMultiTracks(), name="sumTrack_%s_%s" % (name, side))
        sum_multi_tracks.long_name = side + " %s"
        sum_multi_tracks.inputs.out_file = "r-%s_%s.nii.gz" % (name, side)

        if is_invert:
            # NODE 12: Merge tractography and inverted tractography fdt_paths
            merge_paths = Node(Merge(2), name="merge_paths_%s_%s" % (name, side))
            merge_paths.long_name = side + " Direct and inverse tractography merging"
            workflow.connect(probtrackx, "fdt_paths", merge_paths, "in1")
            workflow.connect(probtrackx_inverted, "fdt_paths", merge_paths, "in2")

            # NODE 13: Merge tractography and inverted tractography way_total
            merge_waytotals = Node(
                Merge(2), name="merge_waytotals_%s_%s" % (name, side)
            )
            merge_waytotals.long_name = side + " Direct and inverse waytotal merging"
            workflow.connect(probtrackx, "way_total", merge_waytotals, "in1")
            workflow.connect(probtrackx_inverted, "way_total", merge_waytotals, "in2")

            workflow.connect(merge_paths, "out", sum_multi_tracks, "path_files")
            workflow.connect(merge_waytotals, "out", sum_multi_tracks, "waytotal_files")

        else:
            workflow.connect(probtrackx, "fdt_paths", sum_multi_tracks, "path_files")
            workflow.connect(
                probtrackx, "way_total", sum_multi_tracks, "waytotal_files"
            )

        workflow.connect(
            sum_multi_tracks, "out_file", outputnode, "fdt_paths_%s" % side
        )
        workflow.connect(
            sum_multi_tracks, "waytotal_sum", outputnode, "waytotal_%s" % side
        )

    return workflow



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype import logging as nipype_log, config
import os
import traceback
from multiprocessing import Process, Event
from threading import Thread
from swane.nipype_pipeline.engine.WorkflowReport import WorkflowReport, WorkflowSignals
from nipype.external.cloghandler import ConcurrentRotatingFileHandler
from swane.nipype_pipeline.engine.MonitoredMultiProcPlugin import (
    MonitoredMultiProcPlugin,
)
import logging as orig_log
from swane.nipype_pipeline.MainWorkflow import MainWorkflow
from multiprocessing import Queue


LOG_DIR_NAME = "log"


class WorkflowProcess(Process):
    LOG_CHANNELS = [
        "nipype.workflow",
        "nipype.utils",
        "nipype.filemanip",
        "nipype.interface",
    ]

    def __init__(self, subject_name: str, workflow: MainWorkflow, queue: Queue):
        """
            A Process that execute a subject workflow in a thread, manage executor settings and signaling with gui.
            A process is needed to iterate and kill its subprocess if user wants to stop a workflow, a thread would not
            allow that.

        Parameters
        ----------
        subject_name: str
            The subject name
        workflow: MainWorkflow
            The workflow already generated and populated
        queue: Queue
            The subprocess queue for signal handling

        """
        super(WorkflowProcess, self).__init__()
        self.stop_event: Event = Event()
        self.workflow: MainWorkflow = workflow
        self.queue: Queue = queue
        self.subject_name: str = subject_name

    @staticmethod
    def remove_handlers(handler):
        """
        Remove the specified handler from nipype logger channels
        """

        for channel in WorkflowProcess.LOG_CHANNELS:
            nipype_log.getLogger(channel).removeHandler(handler)

    @staticmethod
    def add_handlers(handler):
        """
        Add the specified handler to nipype logger channels
        """
        for channel in WorkflowProcess.LOG_CHANNELS:
            nipype_log.getLogger(channel).addHandler(handler)

    def workflow_run_worker(self):
        """
        Thread that run the workflow
        """

        plugin_args = {
            "mp_context": "fork",
            "queue": self.queue,
            "status_callback": swane_log_nodes_cb,
        }
        if self.workflow.max_cpu > 0:
            plugin_args["n_procs"] = self.workflow.max_cpu
        if self.workflow.max_gpu > 0:
            plugin_args["n_gpu_proc"] = self.workflow.max_gpu

        try:
            # this is useful to generate resource monitor files in subject directory
            os.chdir(self.workflow.base_dir)

            self.workflow.run(plugin=MonitoredMultiProcPlugin(plugin_args=plugin_args))

        except:
            traceback.print_exc()

        # TODO implement nipype.utils.draw_gantt_chart.generate_gantt_chart but maybe it's bugged

        # This event signal the workflow end. When called here is a finished run
        self.stop_event.set()

    @staticmethod
    def kill_with_subprocess():
        """
        Loop all subprocess of current process and kill them, then kill the parent process.
        MUST BE CALLED FROM THE WORKFLOW PROCESS ITSELF OR WILL KILL THE MAIN APPLICATION!
        """
        import psutil

        try:
            this_process = psutil.Process(os.getpid())
            children = this_process.children(recursive=True)

            for child_process in children:
                try:
                    child_process.kill()
                except psutil.NoSuchProcess:
                    continue
            this_process.kill()
        except psutil.NoSuchProcess:
            return

    def run(self):
        """
        The Process main code
        """
        # log folder management
        log_dir = os.path.join(self.workflow.base_dir, LOG_DIR_NAME)
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        self.workflow.config["execution"]["crashdump_dir"] = log_dir
        self.workflow.config["execution"]["crashfile_format"] = "txt"
        log_filename = os.path.join(log_dir, "pypeline.log")
        file_handler = ConcurrentRotatingFileHandler(
            log_filename,
            maxBytes=int(config.get("logging", "log_size")),
            backupCount=int(config.get("logging", "log_rotate")),
        )
        formatter = orig_log.Formatter(fmt=nipype_log.fmt, datefmt=nipype_log.datefmt)
        file_handler.setFormatter(formatter)
        WorkflowProcess.add_handlers(file_handler)

        # enable resource monitor if required
        if self.workflow.is_resource_monitor:
            config.enable_resource_monitor()
            resource_log_filename = os.path.join(log_dir, "resource_monitor.log")
            callback_logger = orig_log.getLogger("callback")
            callback_logger.setLevel(orig_log.DEBUG)
            resource_log_handler = orig_log.FileHandler(resource_log_filename)
            callback_logger.addHandler(resource_log_handler)

        # start workflow subthread
        workflow_run_work = Thread(target=self.workflow_run_worker)
        workflow_run_work.start()

        # Wait for stop_event. It can be set from the workflow subthread (if workflow is executed till end) or from
        # the GUI to stop the execution
        self.stop_event.wait()

        # Handler removal
        WorkflowProcess.remove_handlers(file_handler)
        if self.workflow.is_resource_monitor:
            callback_logger.removeHandler(resource_log_handler)

        # Signal workflow_stop to GUI and close queue
        self.queue.put(WorkflowReport(signal_type=WorkflowSignals.WORKFLOW_STOP))
        self.queue.close()

        # If the thread is alive at this point the stop_event was set from GUI, so the user asked to kill the process
        if workflow_run_work.is_alive():
            WorkflowProcess.kill_with_subprocess()


# Log node stats function
def swane_log_nodes_cb(node, status):
    """Function to record node run statistics to a log file as json
    dictionaries

    Parameters
    ----------
    node : nipype.pipeline.engine.Node
        the node being logged
    status : string
        acceptable values are 'start', 'end'; otherwise it is
        considered and error

    Returns
    -------
    None
        this function does not return any values, it logs the node
        status info to the callback logger
    """

    if status != "end":
        return

    # Import packages
    import logging
    import json

    status_dict = {
        "name": node.name,
        "id": node._id,
        "start": getattr(node.result.runtime, "startTime", None),
        "finish": getattr(node.result.runtime, "endTime", None),
        "duration": getattr(node.result.runtime, "duration", None),
        "runtime_threads": getattr(node.result.runtime, "cpu_percent", "N/A"),
        "runtime_memory_gb": getattr(node.result.runtime, "mem_peak_gb", "N/A"),
        "estimated_memory_gb": node.mem_gb,
        "num_threads": node.n_procs,
    }

    if status_dict["start"] is None or status_dict["finish"] is None:
        status_dict["error"] = True

    # Dump string to log
    logging.getLogger("callback").debug(json.dumps(status_dict))



from PySide6.QtWidgets import QScrollArea, QGroupBox, QSizePolicy
from PySide6.QtCore import Qt, QEvent


class VerticalScrollArea(QScrollArea):
    """
    Custom implementation of PySide QScrollArea to define a dynamic width scrollable area.

    """

    def __init__(self, parent=None):
        super(VerticalScrollArea, self).__init__(parent)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.m_scrollAreaWidgetContents = QGroupBox(self)
        self.m_scrollAreaWidgetContents.setSizePolicy(
            QSizePolicy.Preferred, QSizePolicy.Preferred
        )
        self.setWidget(self.m_scrollAreaWidgetContents)
        self.m_scrollAreaWidgetContents.installEventFilter(self)

    def eventFilter(self, object, event):
        if object == self.m_scrollAreaWidgetContents and event.type() == QEvent.Resize:
            self.setMinimumWidth(
                self.m_scrollAreaWidgetContents.minimumSizeHint().width()
                + self.verticalScrollBar().width()
            )

        return False



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl.utils import ImageMaths, ImageMathsInputSpec
from nipype.interfaces.base import InputMultiPath, File


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.utils.ImageMathsInputSpec)  -*-
class SumMultiVolsInputSpec(ImageMathsInputSpec):
    vol_files = InputMultiPath(
        File(exists=True), mandatory=True, desc="list of path file to sum togheter"
    )
    in_file = File(exists=True, argstr="%s", position=1)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.utils.ImageMaths)  -*-
class SumMultiVols(ImageMaths):
    """
    Sum multiple volumes.

    """

    input_spec = SumMultiVolsInputSpec

    def _parse_inputs(self, skip=None):
        """
        Custom implementation of _parse_inputs func to manage summ.

        """

        first = True

        for vol in self.inputs.vol_files:
            if first:
                self.inputs.op_string = ""
                self.inputs.in_file = vol
                first = False
            else:
                self.inputs.op_string += "-add " + vol + " "

        parse = super(SumMultiVols, self)._parse_inputs(skip)

        return parse
